# -*- coding: utf-8 -*-

import array
import random
from math import floor, ceil
from collections import namedtuple

from lithoxyl.p_squared import P2Estimator


HistogramCell = namedtuple('HistogramCell',
                           'q_range val_range ratio count')


P2_MIN = (25.0, 50.0, 75.0)
P2_PRAG = (1.0, 5.0, 25.0, 50.0, 75.0, 90.0, 95.0, 99.0)
P2_PRO = (0.1, 1, 2, 5, 10, 25, 50, 75, 80, 85, 90, 95,
          98, 99, 99.5, 99.8, 99.9, 99.99)


class BaseQuantileAccumulator(object):
    def __init__(self):
        self._count = 0
        self._min = float('inf')
        self._max = float('-inf')

    def add(self, val, idx=None):
        self._count += 1
        if val < self._min:
            self._min = val
        if val > self._max:
            self._max = val

    def get_quantiles(self, q_points=None):
        q_points = q_points or P2_PRAG
        ret = [(0.0, self.min)]
        ret.extend([(q, self._get_quantile(q)) for q in q_points])
        ret.append((100.0, self.max))
        return ret

    def get_histogram(self, q_points=None):
        """\
        This convenience method gives back an estimated histogram, based
        on quantiles from get_quantiles(). It's mostly just a utility
        for rendering graphs; it's no more accurate than
        get_quantiles(), and probably even less so for very small
        dataset-size-to-bucket-count ratios.

        TODO: Because it stores observations, this BasicAccumulator
        could actually give back a real histogram, too.
        """
        ret = []
        qwantz = self.get_quantiles(q_points)
        total_count = self.count
        for sq, eq in zip(qwantz, qwantz[1:]):
            q_range = start_q, end_q = sq[0], eq[0]
            val_range = start_val, end_val = sq[1], eq[1]
            ratio = (end_q - start_q) / 100.0
            count = int(ratio * total_count)
            if total_count < len(qwantz):
                if end_val > start_val:
                    count += 1  # not exactly, but sorta.
            else:
                if start_q == 0.0 or end_q == 100.0:
                    count += 1  # make range inclusive
            ret.append(HistogramCell(q_range, val_range, ratio, count))
        return ret

    @property
    def count(self):
        return self._count

    @property
    def min(self):
        return self._min

    @property
    def max(self):
        return self._max

    @property
    def range(self):
        return self._min, self._max

    @property
    def median(self):
        return self._get_quantile(50)

    @property
    def quartiles(self):
        gq = self._get_quantile
        return gq(25), gq(50), gq(75)

    @property
    def iqr(self):
        gq = self._get_quantile
        return gq(75) - gq(25)

    @property
    def trimean(self):
        qs = self.quartiles
        return (qs[0] + (2 * qs[1]) + qs[2]) / 4.0


class ReservoirAccumulator(BaseQuantileAccumulator):
    def __init__(self, data=None, cap=None):
        super(ReservoirAccumulator, self).__init__()
        self._typecode = 'f'  # TODO
        self._data = array.array(self._typecode)
        self._is_sorted = True
        if cap is None:
            self._cap = float('inf')
        elif cap is True:
            self._cap = 2 ** 14
        else:
            self._cap = int(cap)
        data = data or []
        for v in data:
            self.add(v)

    def _sort(self):

        if self._is_sorted:
            return
        if callable(getattr(self._data, 'sort', None)):
            self._data.sort()
        else:
            self._data = array.array(self._typecode, sorted(self._data))
        self._is_sorted = True

    def add(self, val):
        if self._count < self._cap:
            self._data.append(val)
            self._is_sorted = False
            super(ReservoirAccumulator, self).add(val)
        else:
            # TODO: randint has a lot of pure-python arg checking
            # machinery we don't need
            idx = random.randint(0, self._count)
            if idx < self._cap:
                self._data[idx] = val
                self._is_sorted = False
                super(ReservoirAccumulator, self).add(val)

    def _get_quantile(self, q=50):
        if not (0 < q < 100):
            raise ValueError('expected a value in range 0-100 (non-inclusive)')
        self._sort()
        data, n = self._data, len(self._data)
        idx = q / 100.0 * (n - 1)
        idx_f, idx_c = int(floor(idx)), int(ceil(idx))
        if idx_f == idx_c:
            return data[idx_f]
        return (data[idx_f] * (idx_c - idx)) + (data[idx_c] * (idx - idx_f))


class P2Accumulator(BaseQuantileAccumulator):
    def __init__(self, data=None, q_points=P2_PRAG):
        super(P2Accumulator, self).__init__()
        data = data or []
        self._q_points = P2Estimator._process_q_points(q_points)
        self._tmp_acc = ReservoirAccumulator(cap=None)
        self._thresh = len(self._q_points) + 2
        self._est = None

        for v in data:
            self.add(v)

    def add(self, val):
        if self._est is None:
            ta = self._tmp_acc
            ta.add(val)
            if ta.count >= self._thresh:
                self._est = P2Estimator(self._q_points, ta._data)
                self._tmp_acc = None
            return
        else:
            self._est.add(val)
        super(P2Accumulator, self).add(val)

    def get_quantiles(self, q_points=None):
        q_points = q_points or self._q_points
        return super(P2Accumulator, self).get_quantiles(q_points)

    def _get_quantile(self, q):
        try:
            return self._est._get_quantile(q)
        except AttributeError:
            return self._tmp_acc._get_quantile(q)



# -*- coding: utf-8 -*-

import os
import random

from lithoxyl.moment import MomentAccumulator
from lithoxyl.quantile import ReservoirAccumulator, P2Accumulator
import _statsutils


random.seed(8675309)
test_sets = {'urandom 0-255': [ord(x) for x in os.urandom(16000)],
             'random.random 0.0-1.0': [random.random() for i in xrange(100000)]}


def _assert_round_cmp(a, b, mag=3, name=None):
    thresh = 1.0 / (10 ** mag)
    abs_diff = round(abs(a - b), mag + 1)
    tmpl = 'round-compare failed at %d digits (%f - %f = %f > %f)'
    rel_diff = (2 * abs_diff) / (a + b)
    err_msg = tmpl % (mag, a, b, abs_diff, thresh)
    if name:
        err_msg = '%r %s' % (name, err_msg)
    assert rel_diff < thresh, err_msg
    return True


def test_momentacc_basic():
    for name, data in test_sets.items():
        ma = MomentAccumulator()
        for v in data:
            ma.add(v)

        for m_name in ('mean', 'variance', 'std_dev', 'skewness', 'kurtosis'):
            ma_val = getattr(ma, m_name)
            ctl_val = getattr(_statsutils, m_name)(data)
            _assert_round_cmp(ctl_val, ma_val, mag=4, name=m_name)
    return True


def test_momentacc_norm():
    ma = MomentAccumulator()
    for v in [random.gauss(10, 4) for i in xrange(5000)]:
        ma.add(v)
    _assert_round_cmp(10, abs(ma.mean), mag=1)
    _assert_round_cmp(4, ma.std_dev, mag=1)
    _assert_round_cmp(0, ma.skewness, mag=1)
    _assert_round_cmp(3, ma.kurtosis, mag=1)


def test_quantacc_basic(data=None):
    data = data or range(31)
    qa = ReservoirAccumulator()
    for v in data:
        qa.add(v)
    assert qa.median == _statsutils.median(data)
    q1, q2, q3 = qa.quartiles
    assert q1 < q2 < q3
    return True


def test_quantacc():
    for name, data in test_sets.items():
        qa = ReservoirAccumulator()
        for v in data:
            qa.add(v)
        _assert_round_cmp(qa.median, _statsutils.median(data), mag=6)
        q1, q2, q3 = qa.quartiles
        assert q1 < q2 < q3
        hist = qa.get_histogram()
        assert hist
        #print; import pprint; pprint.pprint(hist)
        #print sum([x.count for x in hist]), 'histogram item count'


def test_p2quantacc():
    for name, data in test_sets.items():
        qa = ReservoirAccumulator()
        p2qa = P2Accumulator()
        for i, v in enumerate(data):
            p2qa.add(v)
            qa.add(v)
            if i and i % 1000 == 0:
                _assert_round_cmp(qa.median,
                                  p2qa.median,
                                  mag=1,
                                  name='%s median' % name)
                #print i, qa.median, p2qa.median

        _assert_round_cmp(qa.median,
                          p2qa.median,
                          mag=2,
                          name='%s median' % name)


def test_acc_random():
    data = test_sets['random.random 0.0-1.0']

    qa = ReservoirAccumulator(data)
    capqa = ReservoirAccumulator(data, cap=True)
    p2qa = P2Accumulator(data)
    for acc in (qa, capqa, p2qa):
        for qp, v in acc.get_quantiles():
            if qp > 0:
                assert 0.95 < v / (qp / 100.0) < 1.05



# -*- coding: utf-8 -*-

from lithoxyl import (SensibleSink,
                      SensibleFilter,
                      SensibleFormatter as SF,
                      SensibleMessageFormatter as SMF)
from lithoxyl.emitters import StreamEmitter, AggregateEmitter
from lithoxyl.logger import Logger


fmtr = SF('{status_char}{begin_timestamp}')
strm_emtr = StreamEmitter('stderr')
aggr_emtr = AggregateEmitter()
strm_sink = SensibleSink(formatter=fmtr, emitter=strm_emtr)
fake_sink = SensibleSink(formatter=fmtr, emitter=aggr_emtr)


def test_sensible_basic():
    log = Logger('test_ss', [strm_sink, fake_sink])

    print

    log.debug('greet').success('hey')
    assert aggr_emtr.get_entry(-1).startswith('s')

    with log.debug('greet') as t:
        t.success('hello')
        t.warn("everything ok?")

    assert aggr_emtr.get_entry(-1).startswith('S')

    with log.debug('greet') as t:
        t.failure('bye')
    assert aggr_emtr.get_entry(-1).startswith('F')

    try:
        with log.debug('greet') as t:
            raise ZeroDivisionError('narwhalbaconderp')
    except Exception:
        pass

    assert aggr_emtr.get_entry(-1).startswith('E')


def test_bad_encoding():
    try:
        StreamEmitter('stderr', encoding='nope')
    except LookupError:
        assert True
    else:
        assert False


def test_bad_encoding_error_fallback():
    try:
        StreamEmitter('stderr', errors='badvalue')
    except LookupError:
        assert True
    else:
        assert False


def _test_exception():
    _tmpl = ('{iso_end} - {exc_type}: {exc_message}'
             ' - {func_name}:{line_number} - {exc_tb_list}')
    sink = SensibleSink(SF(_tmpl),
                        StreamEmitter('stderr'),
                        filters=[SensibleFilter(exception=0)])
    logger = Logger('excelsilog', [sink])
    with logger.info('A for Effort', reraise=False) as tr:
        print tr
        raise ValueError('E for Exception')
    return



# -*- coding: utf-8 -*-

import time
from boltons.funcutils import total_ordering


EVENTS = ('begin', 'warn', 'end', 'exception', 'comment')
IMPORT_TIME = time.time()


def to_unicode(obj):
    try:
        return unicode(obj)
    except UnicodeDecodeError:
        return unicode(obj, encoding='utf8', errors='replace')


@total_ordering
class Level(object):
    def __init__(self, name, value):
        self.name = name.lower()
        self._value = value

    def __eq__(self, other):
        if self is other:
            return True
        elif self._value == getattr(other, '_value', None):
            return True
        elif self.name == other:
            return True
        return False

    def __lt__(self, other):
        if self is other:
            return False
        elif self._value < getattr(other, '_value', 100):
            return True
        return False

    def __repr__(self):
        return '%s(%r, %r)' % (self.__class__.__name__, self.name, self._value)


DEBUG = Level('debug', 20)
INFO = Level('info', 70)
CRITICAL = Level('critical', 90)
DEFAULT_LEVEL = DEBUG
BUILTIN_LEVELS = (DEBUG, INFO, CRITICAL)

MIN_LEVEL = Level('_min', 0)
MAX_LEVEL = Level('_max', 100)


def register_level(level_obj):
    if not isinstance(level_obj, Level):
        raise TypeError('expected Level object, not %r' % level_obj)

    LEVEL_ALIAS_MAP[level_obj.name.lower()] = level_obj
    LEVEL_ALIAS_MAP[level_obj.name.upper()] = level_obj
    LEVEL_ALIAS_MAP[level_obj._value] = level_obj
    LEVEL_ALIAS_MAP[level_obj] = level_obj
    LEVEL_LIST[:] = sorted(set(LEVEL_ALIAS_MAP.values()))


LEVEL_LIST = []
LEVEL_ALIAS_MAP = {}
register_level(MIN_LEVEL)
register_level(MAX_LEVEL)
for level in BUILTIN_LEVELS:
    register_level(level)
del level


def get_level(key, default=DEFAULT_LEVEL):
    return LEVEL_ALIAS_MAP.get(key, default)


def get_next_level(key, delta=1):
    level = get_level(key)
    next_i = min(LEVEL_LIST.index(level) + delta, len(LEVEL_LIST) - 1)
    next_level = LEVEL_LIST[next_i]
    return next_level


def get_prev_level(key, delta=1):
    level, delta = get_level(key), abs(delta)
    prev_i = max(LEVEL_LIST.index(level) - delta, 0)
    prev_level = LEVEL_LIST[prev_i]
    return prev_level



# -*- coding: utf-8 -*-
"""Emitters are callable objects which take an entry in *text-form*
and output it to a persistence resource, such as stdout/stderr, files,
or network streams.
"""

import os
import sys
from collections import deque

from lithoxyl.context import note
from lithoxyl.utils import check_encoding_settings

try:
    # unix only
    from lithoxyl._syslog_emitter import SyslogEmitter
except ImportError:
    pass


class AggregateEmitter(object):
    def __init__(self, limit=None):
        self._limit = limit
        self.items = deque(maxlen=limit)

    def get_entries(self):
        return [entry for record, entry in self.items]

    def get_entry(self, idx):
        return self.items[idx][1]

    def clear(self):
        self.items.clear()

    def emit_entry(self, record, entry):
        self.items.append((record, entry))

    on_begin = on_warn = on_end = on_comment = emit_entry

    def __repr__(self):
        cn = self.__class__.__name__
        args = (cn, self._limit, len(self.items))
        msg = '<%s limit=%r entry_count=%r>' % args
        return msg


class StreamEmitter(object):
    def __init__(self, stream, encoding=None, **kwargs):
        if stream == 'stdout':
            stream = sys.stdout
        elif stream == 'stderr':
            stream = sys.stderr
        elif not callable(getattr(stream, 'write', None)):
            raise TypeError('StreamEmitter expected file-like object'
                            ' (or shortcut values "stderr" or "stdout"),'
                            ' not %r' % stream)
        self.stream = stream
        if encoding is None:
            encoding = getattr(stream, 'encoding', None) or 'UTF-8'
        errors = kwargs.pop('errors', 'backslashreplace')

        check_encoding_settings(encoding, errors)  # raises on error

        self.sep = kwargs.pop('sep', None)
        if self.sep is None:
            self.sep = os.linesep
        self.errors = errors
        self.encoding = encoding

    def emit_entry(self, record, entry):
        try:
            entry = entry.encode(self.encoding, self.errors)
        except UnicodeDecodeError as ude:
            # Note that this is a *decode* error, meaning a bytestring
            # found its way through and implicit decoding is happening.
            note('emit_encode', 'got %r on %s.emit_entry();'
                 ' expected decoded text, not %r', ude, self, entry)
            raise
        try:
            self.stream.write(entry)
            if self.sep:
                self.stream.write(self.sep)
            self.flush()
        except Exception as e:
            note('stream_emit', 'got %r on %r.emit_entry()', e, self)

    on_begin = on_warn = on_end = on_comment = emit_entry

    def flush(self):
        stream_flush = getattr(self.stream, 'flush', None)
        if not callable(stream_flush):
            return
        try:
            stream_flush()
        except Exception as e:
            note('stream_flush', 'got %r on %r.flush()', e, self)

    def __repr__(self):
        return '<%s stream=%r>' % (self.__class__.__name__, self.stream)


class FileEmitter(StreamEmitter):
    def __init__(self, filepath, encoding=None, **kwargs):
        self.filepath = os.path.abspath(filepath)
        self.encoding = encoding
        self.mode = 'a'
        stream = open(self.filepath, self.mode)
        super(FileEmitter, self).__init__(stream, self.encoding, **kwargs)

    def close(self):
        if self.stream is None:
            return
        try:
            self.flush()
            self.stream.close()
        except Exception as e:
            note('file_close', 'got %r on %r.close()', e, self)




import math
from bisect import bisect_right, insort

# TODO: trivial to implement reset on these, but recreation might be better practice?

class MaxAccumulator(object):
    # max_list is sorted in ascending order
    def __init__(self, count=10):
        self.max_list = []
        self.max_len = count

    def add(self, val):
        max_list, max_len = self.max_list, self.max_len
        if max_list and val <= max_list[0]:
            return
        if len(max_list) < max_len:
            insort(max_list, val)
            return

        idx = bisect_right(max_list, val)
        if idx == max_len:
            max_list.append(val)
            max_list.pop(0)
        else:
            max_list[idx] = val
        return


class MinAccumulator(object):
    # min_list is sorted in ascending order
    def __init__(self, count=10):
        self.min_list = []
        self.max_len = count

    def add(self, val):
        min_list, max_len = self.min_list, self.max_len
        if min_list and val >= min_list[-1]:
            return
        if len(min_list) < max_len:
            insort(min_list, val)
            return

        idx = bisect_right(min_list, val)
        if idx == 0:
            min_list.insert(0, val)
            min_list.pop()
        else:
            min_list[idx] = val
        return


class HistogramCounter(object):
    def __init__(self, bounds):
        try:
            bounds_float = [float(b) for b in bounds]
        except ValueError:
            raise TypeError('expected iterable of numeric ranges, not %r'
                            % (bounds,))
        if any([math.isnan(b) for b in bounds_float]):
            raise ValueError('expected non-NaN floats, not %r' % (bounds,))
        bounds_set = set(bounds_float)

        bounds_set.add(float('inf'))
        bounds_set.add(float('-inf'))
        bounds_sorted = sorted(bounds_set)
        self.bounds = bounds_sorted
        self.ranges = [(b1, b2) for b1, b2 in
                       zip(bounds_sorted, bounds_sorted[1:])]
        self.buckets = [0] * len(self.ranges)

    def add(self, val):
        bucket_idx = bisect_right(self.bounds, val) - 1
        self.buckets[bucket_idx] += 1
        return

    def get_results(self):
        return zip(self.ranges, self.buckets)



"""\
For the moment, this module is a place to collect patterns, mostly
formats, from existing implementations and standards, so as to make
lithoxyl more immediately useful/usable.
"""

"""\

Lithoxyl is unlikely to go very far beyond providing basic formatting
utilities, most likely a few formatting helpers and format
strings. For the moment, I'm going to use new-style format strings.

Here's a list of values provided by Python's built-in logging library
for creating a formatted message, along with desirability wrt Lithoxyl:

 * args - all the args - no
 * asctime - formatted time - yes, but no. time formats should be more
 explicit
 * created - time.time() of record creation - yes, (Record.begin_time)
 * exc_info - exception tuple or none - yes
 * filename - filename of module issuing the logging call - maybe/probably
 * funcName - name of function making the call - maybe/probably
 * levelname - text version of level - yes
 * levelno - numeric version of level - yes
 * lineno - source line of logging call - maybe/probably
 * module - module from whence logging was called - maybe/probably
 * msecs - millisecond portion of message creation time - no
 * message - formatted message (msg % args) - maybe
 * msg - the format-string form of the message - yes
 * name - logger name - yes
 * pathname - full path name to source file of logging call - maybe/probably
 * process - pid - probably not
 * processName - (obvs) - probably not
 * relativeCreated - time in millisecond between creation and module load - no
 * thread - thread id - probably not
 * threadName - (obvs) - probably not
"""

"""
Apache access log:

  1.202.218.21 - - [22/Jun/2013:06:41:43 -0700] "GET /robots.txt HTTP/1.1" 200 26

  Time format: [day/month/year:hour:minute:second zone]

Nginx default access log:

  78.178.243.200 - - [22/Jun/2013:15:02:31 -0700] "GET /favicon.ico HTTP/1.1" 404 570 "-" "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/27.0.1453.116 Safari/537.36" "-"

  119.63.193.132 - - [22/Jun/2013:14:19:36 -0700] "GET / HTTP/1.1" 200 9755 "-" "Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 6.0)" "-"


Internet time format (RFC3339, ISO8601 compatible):

  1985-04-12T23:20:50.52Z
  1996-12-19T16:39:57-08:00

.. note::

   If the time in UTC is known, but the offset to local time is unknown,
   this can be represented with an offset of "-00:00"

Output of the ``date`` command (aka asctime()):

  Sat Jun 22 15:25:38 PDT 2013

RFC 822 date format (i.e., HTTP headers) (always GMT/UTC):

  Sun, 06 Nov 1994 08:49:37 GMT
"""



# -*- coding: utf-8 -*-

from lithoxyl import DeferredValue
from lithoxyl.logger import Logger, Record
from lithoxyl.sensible import SensibleFormatter as SF



template = ('{status_char}{record_warn_char}{begin_timestamp}'
            ' - {iso_begin_local} - {iso_begin}'
            ' - {logger_name} - {status_str} - {record_name}')


t_log = Logger('1off')
t_riker = Record(t_log, 'DEBUG', 'Riker').success('Hello, Thomas.')

RECS = [t_riker]


TCS = [[('{logger_name}', '"1off"'),
        ('{status_str}', 'success'),
        ('{level_number}', '20'),
        ('{record_name}', '"Riker"'),
        ('{end_message}', '"Hello, Thomas."')]
       ]


def test_formatter_basic():
    forming = SF(template)
    output = forming.on_end(t_riker.end_event)
    expected = '"1off" - success - "Riker"'
    print output
    assert output[-len(expected):] == expected

    rec = Record(t_log, 'DEBUG', 'Wharf')
    robf = SF(template)
    rec.success('Mr. Wolf')
    ret = robf.on_end(rec.end_event)
    print ret
    return ret


def test_individual_fields():
    for record, field_pairs in zip(RECS, TCS):
        for field_tmpl, result in field_pairs:
            forming = SF(field_tmpl)
            output = forming.on_end(record.end_event)
            assert output == result
    return


def test_deferred():
    DV = DeferredValue
    expensive_ops = [(lambda: 5, '"oh, 5"'),
                     (lambda: 'hi', '"oh, hi"'),
                     (lambda: 2.0, '"oh, 2.0"')]
    formatter = SF('{end_message}')

    for eo, expected in expensive_ops:
        rec = Record(t_log, 'DEBUG', 'spendy').success('oh, {dv}', dv=DV(eo))
        output = formatter.on_end(rec.end_event)
        assert output == expected
    return


"""
Formatter tests todos

* no args/kwargs
* sufficient, correct args/kwargs
* anonymous args
* missing kwarg
* missing arg
* type mismatch args
* type mismatch kwargs
* usage of "builtin" fields
"""




from lithoxyl.common import Level, DEBUG, INFO, CRITICAL
from lithoxyl.common import MIN_LEVEL, MAX_LEVEL, DEFAULT_LEVEL
from lithoxyl.common import get_level, get_next_level, get_prev_level


def test_default_level_ordering():
    assert DEBUG != INFO != CRITICAL
    assert MIN_LEVEL < DEBUG < INFO < CRITICAL < MAX_LEVEL
    assert DEBUG <= INFO <= CRITICAL
    assert DEBUG == DEBUG


def test_new_level_equality():
    assert DEBUG == Level('debug', 20)


def test_level_getting():
    assert get_level('debug') == DEBUG
    assert get_level('CRITICAL') == CRITICAL
    assert get_level(int(INFO._value)) == INFO
    assert get_level(DEFAULT_LEVEL) == DEFAULT_LEVEL


def test_level_adjacency():
    assert get_next_level(DEBUG) > DEBUG
    assert get_prev_level(CRITICAL) < CRITICAL



# -*- coding: utf-8 -*-

import sys
import inspect
import itertools


class EncodingLookupError(LookupError):
    pass


class ErrorBehaviorLookupError(LookupError):
    pass


def check_encoding_settings(encoding, errors, reraise=True):
    try:
        # then test error-handler
        u''.encode(encoding)
    except LookupError as le:
        if reraise:
            raise EncodingLookupError(le.message)
        return False
    try:
        # then test error-handler
        # python2.6 does not support kwargs for encode
        u'\xdd'.encode('ascii', errors)
    except LookupError as le:
        if reraise:
            raise ErrorBehaviorLookupError(le.message)
        return False
    except Exception:
        # that ascii encode should never work
        return True


def unwrap(target, attr_name):
    wrapped_func = getattr(target, attr_name)
    try:
        unwrapped_func = wrapped_func.__lithoxyl_wrapped__[3]
    except Exception:
        raise ValueError('%s does not appear to be a wrapped function'
                         % wrapped_func)
    setattr(target, attr_name, unwrapped_func)
    return


def unwrap_all(target=None):
    if target is None or isinstance(target, int):
        target = target or 1
        try:
            target_module_name = sys._getframe(target).f_globals.__name__
            calling_module = sys.modules[target_module_name]
        except Exception:
            raise ValueError('unable to wrap all with target: %r' % target)
        target = calling_module

    for attr_name in dir(target):
        try:
            unwrap(target, attr_name)
        except ValueError:
            continue
    return


def wrap_all(logger, level='info', target=None, skip=None,
             label=None, level_map=None, extras=None):
    """
    """
    ret = []
    extras = extras or {}
    level_map = level_map or {}

    if target is None or isinstance(target, int):
        target = target or 1
        try:
            target_module_name = sys._getframe(target).f_globals.__name__
            calling_module = sys.modules[target_module_name]
        except Exception:
            raise ValueError('unable to wrap all with target: %r' % target)
        target = calling_module

    if skip is None:
        skip = '_'
    if isinstance(skip, basestring):
        skip_func = lambda attr_name: skip and attr_name.startswith(skip)
    elif callable(skip):
        skip_func = skip
    else:
        raise ValueError('skip expected string prefix or callable, not %r' %
                         (skip,))

    if label is None:
        try:
            label = target.__name__
        except AttributeError:
            label = '(%s@%s)' % (target.__class__.__name__, hex(id(target)))
    label = str(label)

    for attr_name in dir(target):
        if skip_func(attr_name):
            continue
        val = getattr(target, attr_name)
        if not callable(val) or isinstance(val, type):
            continue
        kwargs = dict(extras)

        kwargs['level'] = level_map.get(attr_name, level)
        kwargs['record_name'] = label + '.' + attr_name

        log_wrapper = logger.wrap(**kwargs)

        wrapped_func = log_wrapper(val)
        setattr(target, attr_name, wrapped_func)
        ret.append(attr_name)

    return ret


"""decorator.py is bad because it excessively changes your decorator
API to be reliant on decorator.py's strange aesthetic. A pre-existing
decorator can't easily be migrated, and a decorator.py decorator is
not compatible with functools.wraps.

Function signature propagation is orthogonal to closure usage. The
author of decorator.py seems to find a problem with having a function
inside of a function and/or relying on closures and/or functools.wraps
interface.
"""



# -*- coding: utf-8 -*-

import time

from lithoxyl.logger import Logger
from lithoxyl.sinks import RateSink, RateAccumulator


def test_rate_sink():
    sink = RateSink()
    logger = Logger('testlog', sinks=[sink])

    for i in range(10):
        with logger.info('sleeping', reraise=False):
            time.sleep(0.02)
            if i % 2:
                raise ValueError()
    test_rates = sink.get_rates()['testlog']['sleeping']
    assert 48 <= round(test_rates['__all__']) <= 51
    assert 24 <= round(test_rates['exception']) <= 26

    counts = sink.get_total_counts()
    assert counts['__all__'] == 10

    assert repr(sink)


def test_rate_acc():
    acc = RateAccumulator()
    target_rate = 2000
    incr = 1.0 / target_rate
    start_time = cur_time = time.time()

    for i in range(1000):
        acc.add(cur_time)
        cur_time = start_time + (incr * i)

    rate = acc.get_rate(start_time=start_time, end_time=cur_time)
    assert target_rate - 2.5 <= rate <= target_rate + 2.5

    new_rate = acc.get_rate(start_time=start_time + 0.5, end_time=cur_time)
    assert round(new_rate - rate, -2) == 0
    # TODO: check the math on the above. It's the "if not count" block
    # in the RateAccumulator. might have an off-by-one

    assert repr(acc)



# -*- coding: utf-8 -*-

import os
import sys
import atexit
import signal

from lithoxyl.actors import IntervalThreadActor

DEFAULT_JOIN_TIMEOUT = 0.5

LITHOXYL_CONTEXT = None


def get_context():
    if not LITHOXYL_CONTEXT:
        set_context(LithoxylContext())

    return LITHOXYL_CONTEXT


def set_context(context):
    global LITHOXYL_CONTEXT

    LITHOXYL_CONTEXT = context

    return context


def note(name, message, *a, **kw):
    return get_context().note(name, message, *a, **kw)


_CONSEC_ACTIVE_REC_MAP = {}
# TODO: idea to be implemented elsewhere: a map of parent records to
# lists of children, to allow children to look up their siblings. will
# allow properly-ordered, non-overlapping output in many concurrent
# applications


# consec = consecutive. this approach works in systems without
# concurrency. threads, gevent, etc. will require separate hooks.
def _consec_get_active_parent(logger, record):
    return _CONSEC_ACTIVE_REC_MAP.get(logger)


def _consec_set_active_parent(logger, record):
    # record can be None to unset the oldest record
    if record:
        _CONSEC_ACTIVE_REC_MAP[logger] = record
    else:
        _CONSEC_ACTIVE_REC_MAP.pop(logger, None)
    return


class LithoxylContext(object):
    def __init__(self, **kwargs):
        self.loggers = []  # TODO: weakref?

        self.async_mode = False
        self.async_actor = None
        self.async_timeout = DEFAULT_JOIN_TIMEOUT
        self._async_atexit_registered = False

        self.note_handlers = []

        self.get_active_parent = kwargs.pop('get_active_parent',
                                            _consec_get_active_parent)
        self.set_active_parent = kwargs.get('set_active_parent',
                                            _consec_set_active_parent)

    def note(self, name, message, *a, **kw):
        """Lithoxyl can't use itself internally. This is a hook for recording
        all of those error conditions that need to be robustly
        ignored, such as if a user's message doesn't format correctly
        at runtime.
        """
        if not self.note_handlers:
            return
        if a:
            try:
                message = message % a
            except Exception:
                pass
        for nh in self.note_handlers:
            nh(name, message)
        return

    def enable_async(self, **kwargs):
        update_loggers = kwargs.pop('update_loggers', True)
        update_sigterm = kwargs.pop('update_sigterm', True)
        update_actor = kwargs.pop('update_actor', True)
        actor_kw = {'task': self.flush,
                    'interval': kwargs.pop('interval', None),
                    'max_interval': kwargs.pop('max_interval', None),
                    # be very careful when not daemonizing thread
                    'daemonize_thread': kwargs.pop('daemonize_thread', True),
                    'note': self.note}
        if kwargs:
            raise TypeError('unexpected keyword arguments: %r' % kwargs.keys())

        self.async_mode = True

        if update_sigterm:
            install_sigterm_handler()

        if update_actor:
            if not self.async_actor:
                self.async_actor = IntervalThreadActor(**actor_kw)
            self.async_actor.start()

        if update_loggers:
            for logger in self.loggers:
                logger.set_async(False)

        # graceful thread shutdown and sink flushing
        if not self._async_atexit_registered:
            # disable_async is safe to call multiple times but this is cleaner
            atexit.register(self.disable_async)
            self._async_atexit_registered = True

        return

    def disable_async(self, **kwargs):
        update_loggers = kwargs.pop('update_loggers', True)
        update_sigterm = kwargs.pop('update_sigterm', True)
        update_actor = kwargs.pop('update_actor', True)
        join_timeout = kwargs.pop('join_timeout', self.async_timeout)

        if update_sigterm:
            uninstall_sigterm_handler()

        if update_actor and self.async_actor:
            self.async_actor.stop()
            self.async_actor.join(join_timeout)

        if update_loggers:
            for logger in self.loggers:
                logger.set_async(False)

        self.flush()
        self.async_mode = False

    def flush(self):
        for logger in self.loggers:
            try:
                logger.flush()
            except Exception as e:
                self.note('context_flush',
                          'exception during logger (%r) flush: %r', logger, e)
        return

    def add_logger(self, logger):
        if logger not in self.loggers:
            self.loggers.append(logger)

    def remove_logger(self, logger):
        try:
            self.loggers.remove(logger)
        except ValueError:
            pass


def signal_sysexit(signum, frame):
    # return code ends up being 143 for sigterm, See page 544 Kerrisk
    # for more see atexit_reissue_sigterm docstring for more details
    atexit.register(atexit_reissue_sigterm)
    sys.exit(143)  # approximate sigterm (128 + 15)


def atexit_reissue_sigterm():
    """The only way to "transparently" handle SIGTERM and terminate with
    the same status code as if we did not have a handler installed is
    to uninstall the handler and reissue the SIGTERM signal. Kerrisk
    p549 details more.

    So our signal handler registers this atexit handler, which calls
    :func:`os.kill`.

    Because that will end the process, extra precautions are built in
    to make sure we are the last exit handler running.

    """
    # best attempt at ensuring that we run last
    global _ASYNC_ATEXIT_ATTEMPT_LAST
    try:
        func, _, _ = atexit._exithandlers[0]
        if func is not atexit_reissue_sigterm and _ASYNC_ATEXIT_ATTEMPT_LAST:
            _ASYNC_ATEXIT_ATTEMPT_LAST = False
            atexit._exithandlers.insert(0, (atexit_reissue_sigterm, (), {}))
            return
    except IndexError:
        pass  # _exithandlers is empty, this is the last exitfunc
    except Exception:
        # TODO: effing atexit runs exitfuncs LIFO. If we os.exit early,
        # it's less grace, so abort reissuing the signal
        return

    uninstall_sigterm_handler(force=True)
    os.kill(os.getpid(), 15)
    return


_ASYNC_ATEXIT_ATTEMPT_LAST = True


def install_sigterm_handler():
    """This installs a no-op Python SIGTERM handler to ensure that atexit
    functions are called. If there is already a SIGTERM handler, no
    new handler is installed.
    """
    cur = signal.getsignal(signal.SIGTERM)
    if cur == signal.SIG_DFL:
        signal.signal(signal.SIGTERM, signal_sysexit)
        return True
    return False


def uninstall_sigterm_handler(force=False):
    cur = signal.getsignal(signal.SIGTERM)
    if force or cur is signal_sysexit:
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        return True
    return False



# -*- coding: utf-8 -*-


class MomentAccumulator(object):
    """\
    An accumulator for tracking statistical moments. Supports
    arithmetic mean, variance, standard deviation, and the
    distribution shape-related moments, skewness and kurtosis.

    It operates using online approaches, and does not store
    observations. As a result, it uses a more-or-less fixed amount of
    memory, but has a higher insertion cost. Also, because this is
    using simple Python doubles, it is likely to build up a good
    amount of floating point drift. If that's not good enough, use an
    accumulator that stores observations, make a Decimal version, or
    use NumPy, I guess.

    Almost called it the KodakAccumulator, but trademarks, bad jokes,
    etc.

    N.B. For convenience, values default to -0.0 when initially
    uncomputable, such as mean when there is no data, or variance when
    there's only one data point. I would use NaN, but in Python (2 and
    3), NaN's repr() does not round-trip. I might revisit this
    interface optimization later.
    """
    def __init__(self):
        self._count = 0
        self._min = float('inf')
        self._max = float('-inf')
        self._mean = -0.0
        self._m2 = -0.0
        self._m3 = -0.0
        self._m4 = -0.0

    def add(self, val):
        # TODO: keep max/min? not strictly necessary here.
        if val > self._max:
            self._max = val
            if self._count == 0:
                self._min = val
        elif val < self._min:
            self._min = val

        self._count += 1
        n, m2, m3, m4 = self._count, self._m2, self._m3, self._m4
        delta = val - self._mean
        delta_n = delta / n
        delta_n2 = delta_n ** 2
        term = delta * delta_n * (n - 1)
        self._mean = self._mean + delta_n
        self._m4 = (m4 +
                    term * delta_n2 * (n ** 2 - 3 * n + 3) +
                    6 * delta_n2 * m2 -
                    4 * delta_n * m3)
        self._m3 = (m3 +
                    term * delta_n * (n - 2) -
                    3 * delta_n * m2)
        self._m2 = m2 + term

    @property
    def count(self):
        return self._count

    @property
    def mean(self):
        return self._mean

    @property
    def variance(self):
        try:
            return self._m2 / (self._count - 1)
        except ArithmeticError:
            return -0.0

    @property
    def skewness(self):
        try:
            return ((self._count ** 0.5) * self._m3) / (self._m2 ** 1.5)
        except ArithmeticError:
            return -0.0

    @property
    def kurtosis(self):
        # TODO: subtract 3? (for normal curve = 0)
        try:
            return (self._count * self._m4) / (self._m2 ** 2)
        except ArithmeticError:
            return -0.0

    @property
    def std_dev(self):
        return self.variance ** 0.5



# -*- coding: utf-8 -*-
"""\

An implementation of P-Squared (Piecewise Parabolic) Quantile
Estimation, which provides efficient online estimation for
quantile-related statistics (e.g., median, quartiles).

For description of the algorithm defined in
http://www.cs.wustl.edu/~jain/papers/ftp/psqr.pdf

Implemented by Kurt Rose and Mahmoud Hashemi.

Copyright 2013, 3-clause BSD License
"""


class P2Estimator(object):
    def __init__(self, q_points, data):
        self._q_points = self._process_q_points(q_points)
        self._q_points = (0.0,) + self._q_points + (100.0,)
        len_data, len_qps = len(data), len(self._q_points)
        if len_data < len_qps:
            msg = ('expected %d or more initial points for '
                   '%d quantiles (got %d)' % (len_qps, len_qps - 2, len_data))
            raise ValueError(msg)

        initial = sorted(data[:len_qps])
        vals = [[i + 1, x] for i, x in enumerate(initial)]
        self._points = pts = zip(self._q_points, vals)
        self._min_point, self._max_point = pts[0][1], pts[-1][1]
        self._lookup = dict(pts)
        self._back_tuples = list(reversed(zip(vals[1:], vals[2:])))
        qp_divs = [q / 100.0 for q in self._q_points]
        self._quads = zip(qp_divs[1:], vals, vals[1:], vals[2:])

        for i in xrange(len_qps, len_data):
            self.add(data[i])

    @staticmethod
    def _process_q_points(q_points):
        try:
            qps = sorted([float(x) for x in set(q_points or [])])
            if qps[0] == 0.0:
                qps = qps[1:]
            if qps[-1] == 100.0:
                qps = qps[:-1]
            if not qps or not all([0 < x < 100 for x in qps]):
                raise ValueError()
        except Exception:
            raise ValueError('invalid quantile point(s): %r' % (q_points,))
        else:
            return tuple(qps)

    def add(self, val):
        prev_count = self._max_point[0]
        self._max_point[0] = prev_count + 1

        cur_min, cur_max = self._min_point[1], self._max_point[1]
        if val < cur_min:
            self._min_point[1] = cur_min = val
        elif val > cur_max:
            self._max_point[1] = cur_max = val

        for point, nxt_point in self._back_tuples:
            if val <= point[1]:
                point[0] += 1
                if point[0] == nxt_point[0]:
                    point[0] -= 1

        # update estimated locations of percentiles
        for qpdiv, (ln, lq), cur, (rn, rq) in self._quads:
            cn, cq = cur
            d = int(prev_count * qpdiv + 1 - cn)
            if not d:
                continue
            d = 1.0 if d > 0 else -1.0  # clamped at +-1
            if not (ln < cn + d < rn):
                continue
            nq = (cq + (d / (rn - ln)) *  # hooray parabolic
                  ((cn - ln + d) * (rq - cq) / (rn - cn) +
                   (rn - cn - d) * (cq - lq) / (cn - ln)))
            if not (lq < nq < rq):  # fall back on linear eqn
                if d == 1:
                    nq = cq + (rq - cq) / (rn - cn)
                else:
                    nq = cq - (lq - cq) / (ln - cn)
            cur[0], cur[1] = cn + d, nq

    def get_quantiles(self):
        return [(x[0], x[1][1]) for x in self._points]

    def _get_quantile(self, q):
        try:
            return self._lookup[float(q)][1]
        except KeyError:
            raise ValueError('quantile not tracked: %r' % q)


def test_random(vals=None, nsamples=100000):
    import random
    import time
    from pprint import pprint
    random.seed(12345)
    qp = (1.0, 5.0, 25.0, 50.0, 75.0, 90.0, 95.0, 99.0)
    if not vals:
        vals = [random.random() for i in range(nsamples)]
    try:
        start = time.time()
        m = P2Estimator(q_points=qp, data=vals)
        p = m.get_quantiles()
        duration = time.time() - start
        tmpl = ("P2Estimator processed %d measurements "
                "in %f seconds (%f ms each)")
        pprint(p)
        print tmpl % (nsamples, duration, 1000 * duration / nsamples)
    except Exception:
        import traceback
        import pdb
        traceback.print_exc()
        pdb.post_mortem()
        raise
    for k, v in p:
        if not k:
            continue
        if not 0.95 < v / (k / 100.0) < 1.05:
            print "problem: %s is %s, should be ~%s" % (k, v, k / 100.0)
    return


if __name__ == "__main__":
    test_random()



# -*- coding: utf-8 -*-

import time

from lithoxyl.logger import Logger
from lithoxyl.context import get_context, LithoxylContext


def test_async_on_off():
    ctx = get_context()

    ctx.enable_async()
    time.sleep(0.25)  # 250ms should be plenty

    assert ctx.async_actor.is_alive()

    ctx.disable_async()
    time.sleep(0.1)  # 100ms should be plenty

    assert not ctx.async_actor.is_alive()

    return


def test_preflush_catching_and_noting():
    ctx = LithoxylContext()

    def raiser(log):
        raise RuntimeError('never gonna catch me')

    log = Logger('test_logger', context=ctx)
    log.preflush_hooks.append(raiser)

    notes = []

    def add_note(name, message):
        notes.append((name, message))

    ctx.note_handlers.append(add_note)

    ctx.enable_async()
    time.sleep(0.3)

    assert notes  # should have at least one note in 300ms



# -*- coding: utf-8 -*-

import pdb
import time
import bisect
from collections import deque

from boltons.cacheutils import ThresholdCounter as TCounter

from lithoxyl.emitters import StreamEmitter
from lithoxyl.quantile import ReservoirAccumulator, P2Accumulator
from lithoxyl.ewma import EWMAAccumulator, DEFAULT_PERIODS, DEFAULT_INTERVAL


class AggregateSink(object):
    "A simple sink that just aggregates the messages."
    def __init__(self, limit=None):
        self._limit = limit
        self.begin_events = deque(maxlen=limit)
        self.warn_events = deque(maxlen=limit)
        self.end_events = deque(maxlen=limit)
        self.comment_events = deque(maxlen=limit)

    def on_begin(self, begin_event):
        self.begin_events.append(begin_event)

    def on_warn(self, warn_event):
        self.warn_events.append(warn_event)

    def on_end(self, end_event):
        self.end_events.append(end_event)

    def on_comment(self, comment_event):
        self.comment_events.append(comment_event)

    def __repr__(self):
        cn = self.__class__.__name__
        lens = (self._limit, len(self.begin_events), len(self.end_events),
                len(self.warn_events), len(self.comment_events))
        args = (cn,) + lens
        msg = '<%s limit=%r begins=%r ends=%r warns=%r comments=%r>' % args
        return msg


class DevDebugSink(object):
    """Use this to insert debug prompts where exceptions are raised. Pass
    the exception type or iterable of exception types to selectively
    match. Alternatively, pass True for all Exception subtypes.

    """
    def __init__(self, reraise=False, post_mortem=False):
        if reraise is True:
            reraise = Exception
        self.reraise = reraise
        if post_mortem is True:
            post_mortem = Exception
        self.post_mortem = post_mortem

    def on_exception(self, event, exc_type, exc_obj, exc_tb):
        if self.post_mortem and isinstance(exc_obj, self.post_mortem):
            pdb.post_mortem()
        if self.reraise and isinstance(exc_obj, self.reraise):
            raise exc_type, exc_obj, exc_tb
        return


class RateAccumulator(object):
    """\
    The RateAccumulator provides basic accounting and rate estimation
    capabilities based on a single stream of timestamps. Note that the stream
    is assumed to be in order.
    """
    def __init__(self, sample_size=128):
        self.times = deque(maxlen=sample_size)
        self.total_count = 0
        self.creation_time = time.time()

    @property
    def sample_size(self):
        return self.times.maxlen

    def add(self, timestamp):
        """
        Adds a timestamp to the accumulator. Note that timestamps are
        expected to be added _in order_.
        """
        self.total_count += 1
        self.times.append(timestamp)

    def get_norm_times(self, ndigits=4):
        """\
        Mostly for debugging: returns the current reservoir of
        timestamps normalized to the first.
        """
        if not self.times:
            return []
        first = self.times[0]
        return [round(x - first, ndigits) for x in self.times]

    def get_rate(self, start_time=None, end_time=None):
        """\ Returns the per-second rate of the accumulator, taking into
        account the window designated by start_time and
        end_time.

        start_time defaults to the creation time of the
        accumulator. end_time defaults to current time.

        Note that if the window extends beyond the time range
        currently tracked by the reservoir, but less than the total
        lifespan of the accumulator, the rate will be an estimate.
        """
        if not self.times:
            return 0.0
        end_time = end_time or time.time()
        start_time = start_time or self.creation_time
        if start_time <= self.creation_time:
            count = self.total_count
        else:
            target_idx = bisect.bisect_left(self.times, start_time)
            count = len(self.times) - target_idx
            if not count:
                # if our reservoir has nothing to offer, i.e., the
                # rate is high enough to have lost all real datapoints
                # for the window described, we just return the current
                # rate of everything in the reservoir.
                count = len(self.times)
                start_time, end_time = self.times[0], self.times[-1]
            elif not target_idx:
                start_time = self.times[0]
        return count / (end_time - start_time)

    def __repr__(self):
        cn = self.__class__.__name__
        rate = self.get_rate()
        return '<%s rate=%.4f count=%r>' % (cn, rate, self.total_count)


class RateSink(object):
    """\
    The RateSink provides basic accounting and rate estimation
    facilities for records as they pass through the system.

    It uses a reservoir system for predictable and stable memory
    use. See RateAccumulator for more information.

    A RateSink can be shared across multiple loggers.
    """
    def __init__(self, getter=None, sample_size=128):
        if getter is None:
            getter = lambda end_event: end_event.etime
        if not callable(getter):
            raise TypeError('expected callable getter, not %r' % getter)
        self.getter = getter
        self.acc_map = {}
        self.sample_size = sample_size
        self.creation_time = time.time()

    def on_end(self, end_event):
        ev = end_event
        name_time_map = self.acc_map.setdefault(ev.logger, {})
        status_time_map = name_time_map.setdefault(ev.name, {})
        try:
            acc = status_time_map[ev.status]
        except Exception:
            acc = RateAccumulator(sample_size=self.sample_size)
            status_time_map[ev.status] = acc
        acc.add(self.getter(ev))

    def get_rates(self, max_time=None, **kw):
        """\
        Gets a dictionary of rates, grouped by logger and
        status. Aggregates are put under the special key,
        '__all__'. All rates are in terms of seconds.

        The caller can specify `start_time` and `end_time`, or use the
        convenience parameter `max_time`, which specifies how long
        before the current time the window should extend.
        """
        end_time = kw.pop('end_time', time.time())
        start_time = kw.pop('start_time', self.creation_time)
        if max_time:
            start_time = end_time - max_time

        ret = {}
        all_loggers_rate = 0.0
        for logger, name_map in self.acc_map.items():
            cur_logger_rate = 0.0
            ret[logger.name] = {}
            for name, status_map in name_map.items():
                cur_name_rate = 0.0
                ret[logger.name][name] = {}
                for status, acc in status_map.items():
                    cur_rate = acc.get_rate(start_time=start_time,
                                            end_time=end_time)
                    ret[logger.name][name][status] = cur_rate
                    cur_name_rate += cur_rate
                    cur_logger_rate += cur_rate
                    all_loggers_rate += cur_rate
                ret[logger.name][name]['__all__'] = cur_name_rate
            ret[logger.name]['__all__'] = cur_logger_rate
        ret['__all__'] = all_loggers_rate
        return ret

    def get_total_counts(self):
        """\
        Gets a dictionary of counts, grouped by logger and
        status. Aggregates are put under the special key,
        '__all__'. All counts are for the full lifetime of the sink.
        """
        ret = {}
        all_loggers_count = 0
        for logger, name_map in self.acc_map.items():
            cur_logger_count = 0
            ret[logger.name] = {}
            for name, status_map in name_map.items():
                cur_name_count = 0
                ret[logger.name][name] = {}
                for status, acc in status_map.items():
                    cur_count = acc.total_count
                    ret[logger.name][name][status] = cur_count
                    cur_name_count += cur_count
                    cur_logger_count += cur_count
                    all_loggers_count += cur_count
                ret[logger.name][name]['__all__'] = cur_name_count
            ret[logger.name]['__all__'] = cur_logger_count
        ret['__all__'] = all_loggers_count
        return ret

    def __repr__(self):
        cn = self.__class__.__name__
        total_rate_all = self.get_rates()['__all__']
        total_count_all = self.get_total_counts()['__all__']
        args = (cn, total_rate_all, total_count_all)
        return '<%s total_rate=%.4f total_count=%r>' % args


class EWMASink(object):
    accumulator_type = EWMAAccumulator

    def __init__(self,
                 getter=None,
                 periods=DEFAULT_PERIODS,
                 interval=DEFAULT_INTERVAL):
        if getter is None:
            getter = lambda record: record.duration
        if not callable(getter):
            raise TypeError('expected callable getter, not %r' % getter)
        self.getter = getter
        self.periods = periods
        self.interval = interval
        self.acc_map = {}

    def on_end(self, record):
        name_time_map = self.acc_map.setdefault(record.logger.name, {})
        status_time_map = name_time_map.setdefault(record.name, {})
        try:
            acc = status_time_map[record.status]
        except Exception:
            acc = self.accumulator_type(periods=self.periods,
                                        interval=self.interval)
            status_time_map[record.status] = acc
        value = self.getter(record)
        acc.add(value)

    @staticmethod
    def _update_add(target, other):
        for k, v in other.iteritems():
            if v is None:
                continue
            try:
                target[k] += v
            except KeyError:
                target[k] = v
        return

    def get_values(self):
        ret = {}
        _update_add = self._update_add
        all_loggers_vals = ret['__all__'] = {}
        for logger, name_map in self.acc_map.items():
            ret[logger] = {}
            cur_logger_vals = ret[logger]['__all__'] = {}
            for name, status_map in name_map.items():
                ret[logger][name] = {}
                cur_name_vals = ret[logger][name]['__all__'] = {}
                for status, acc in status_map.items():
                    cur_vals = acc.get_rates()
                    ret[logger][name][status] = cur_vals
                    _update_add(cur_name_vals, cur_vals)
                    _update_add(cur_logger_vals, cur_vals)
                    _update_add(all_loggers_vals, cur_vals)
        return ret

    def __repr__(self):
        cn = self.__class__.__name__
        total_values = self.get_values()['__all__']
        return '<%s total_values=%r>' % (cn,
                                         total_values)


class QuantileSink(object):
    def __init__(self, getter=None, **kwargs):
        """There are two approaches for quantile-based stats
        accumulation. A standard, reservoir/replacement strategy
        (QuantileAccumulator) and the P2 approach
        (P2QuantileAccumulator).

        P2 is slower to update, but faster to read. so consider
        setting use_p2 to True if your use case entails more frequent
        stats reading. P2 is also more space-efficient and tends to be
        more accurate for common performance curves.
        """
        if getter is None:
            getter = lambda event: event.duration
        if not callable(getter):
            raise TypeError('expected callable getter, not %r' % getter)
        self.getter = getter

        default_acc = ReservoirAccumulator
        use_p2 = kwargs.pop('use_p2', False)
        if use_p2:
            default_acc = P2Accumulator

        acc_type = kwargs.pop('acc_type', default_acc)

        self._acc_type = acc_type

        self.qas = {}

    def to_dict(self):
        ret = {}
        for r_name, acc in self.qas.iteritems():
            ret[r_name] = {'count': acc.count,
                           'trimean': acc.trimean,
                           'quantiles': dict(acc.get_quantiles())}
        return ret

    def on_end(self, event):
        try:
            logger_accs = self.qas[event.logger.name]
        except KeyError:
            logger_accs = self.qas[event.logger.name] = {}
        try:
            acc = logger_accs[event.name]
        except KeyError:
            acc = logger_accs[event.name] = self._acc_type()

        acc.add(self.getter(event))

    def __repr__(self):
        cn = self.__class__.__name__
        acc_dict_repr = dict([(lname,
                               dict([(k, (a.count, round(a.median, 4)))
                                     for k, a in a_map.items()]))
                              for lname, a_map in self.qas.items()])
        ret = '<%s %r>' % (cn, acc_dict_repr)
        return ret


class CounterSink(object):
    def __init__(self, getter=None, threshold=0.001):
        if getter is None:
            getter = lambda end_event: end_event.record.name
        if not callable(getter):
            raise TypeError('expected callable getter, not %r' % (getter,))

        self.getter = getter
        self.threshold = threshold
        self.counter_map = {}

    def on_end(self, end_event):
        ev, ctr_map = end_event, self.counter_map
        try:
            counter = ctr_map[ev.record.logger]
        except KeyError:
            counter = ctr_map[ev.record.logger] = TCounter(self.threshold)

        key = self.getter(end_event)
        counter.add(key)
        return

    def to_dict(self):
        ret = {}
        for logger, counter in self.counter_map.items():
            ret[logger.name] = cur = dict(counter)
            uncommon_count = counter.get_uncommon_count()
            if uncommon_count:
                ret[logger.name]['__missing__'] = uncommon_count
            cur['__all__'] = sum(cur.values())
        return ret



# -*- coding: utf-8 -*-

import sys
import time
import itertools

from boltons.tbutils import ExceptionInfo, Callpoint

from lithoxyl.context import note
from lithoxyl.sensible import SensibleMessageFormatter
from lithoxyl.common import to_unicode, get_level


_REC_ID_ITER = itertools.count()


class DefaultException(Exception):
    "Only used when traceback extraction fails"


class Record(object):
    """The Record type is one of the core Lithoxyl types, and the key to
    instrumenting application logic. Records are usually instantiated
    through convenience methods on :class:`~lithoxyl.logger.Logger`
    instances, associated with their level (e.g.,
    :meth:`~Logger.critical`).

    Args:
        logger: The Logger instance responsible for creating and
            publishing the Record.
        level: Log level of the Record. Generally one of
            :data:`~lithoxyl.common.DEBUG`,
            :data:`~lithoxyl.common.INFO`, or
            :data:`~lithoxyl.common.CRITICAL`. Defaults to ``None``.
        name (str): A string description of some application action.
        data (dict): A mapping of non-builtin fields to user
            values. Defaults to an empty dict (``{}``) and can be
            populated after Record creation by accessing the Record
            like a ``dict``.
        reraise (bool): Whether or not the Record should catch and
            reraise exceptions. Defaults to ``True``. Setting to
            ``False`` will cause all exceptions to be caught and
            logged appropriately, but not reraised. This should be
            used to eliminate ``try``/``except`` verbosity.
        frame: Frame of the callpoint creating the Record. Defaults to
            the caller's frame.

    Most of these parameters are managed by the Records and respective
    :class:`~lithoxyl.Logger` themselves. While they are provided here
    for advanced use cases, usually only the *name* and *raw_message*
    are provided.

    Records are :class:`dict`-like, and can be accessed as mappings

    and used to store additional structured data:

    >>> record['my_data'] = 20.0
    >>> record['my_lore'] = -record['my_data'] / 10.0
    >>> from pprint import pprint
    >>> pprint(record.data_map)
    {'my_data': 20.0, 'my_lore': -2.0}

    """
    _is_trans = None
    _defer_publish = False

    def __init__(self, logger, level, name,
                 data=None, reraise=True, parent=None, frame=None):
        self.record_id = next(_REC_ID_ITER)
        self.logger = logger
        self.level = get_level(level)
        self.name = name

        self.data_map = data or {}
        self._reraise = reraise

        if frame is None:
            frame = sys._getframe(1)
        self.callpoint = Callpoint.from_frame(frame)

        self.begin_event = None
        self.end_event = None
        # these can go internal and be lazily created through properties
        self.warn_events = []
        self.exc_events = []

        if parent:
            self.parent_record = parent
        else:
            self.parent_record = logger.context.get_active_parent(logger, self)
        return

    def __repr__(self):
        cn = self.__class__.__name__
        return ('<%s %r %s %r>'
                % (cn, self.name, self.level.name.upper(), self.status))

    @property
    def level_name(self):
        return self.level.name

    @property
    def status(self):
        try:
            return self.end_event.status
        except AttributeError:
            return 'begin'

    @property
    def duration(self):
        try:
            return self.end_event.etime - self.begin_event.etime
        except Exception:
            return 0.0

    @property
    def parent_depth(self):
        i = 0
        while self.parent_record and i < 500:
            i += 1
            self = self.parent_record  # reuse var
        return i

    def begin(self, message=None, *a, **kw):
        self.data_map.update(kw)
        if not self.begin_event:
            if not message:
                message = self.name + ' beginning'

            self.begin_event = BeginEvent(self, time.time(), message, a)
            self.logger.on_begin(self.begin_event)
        return self

    def warn(self, message, *a, **kw):
        self.data_map.update(kw)
        warn_ev = WarningEvent(self, time.time(), message, a)
        self.warn_events.append(warn_ev)
        self.logger.on_warn(warn_ev)
        return self

    def success(self, message=None, *a, **kw):
        """Mark this Record successful. Also set the Record's
        *message* template. Positional and keyword arguments will be
        used to generate the formatted message. Keyword arguments will
        also be added to the Record's ``data_map`` attribute.
        """
        if not message:
            if self.data_map:
                message = self.name + ' succeeded - ({data_map_repr})'
            else:
                message = self.name + ' succeeded'
        return self._end('success', message, a, kw)

    def failure(self, message=None, *a, **kw):
        """Mark this Record failed. Also set the Record's
        *message* template. Positional and keyword arguments will be
        used to generate the formatted message. Keyword arguments will
        also be added to the Record's ``data_map`` attribute.
        """
        if not message:
            if self.data_map:
                message = self.name + ' failed - ({data_map_repr})'
            else:
                message = self.name + ' failed'

        return self._end('failure', message, a, kw)

    def exception(self, message=None, *a, **kw):
        """Mark this Record as having had an exception. Also
        sets the Record's *message* template similar to
        :meth:`Record.success` and :meth:`Record.failure`.

        Unlike those two attributes, this method is rarely called
        explicitly by application code, because the context manager
        aspect of the Record catches and sets the appropriate
        exception fields. When called explicitly, this method should
        only be called in an :keyword:`except` block.
        """
        exc_type, exc_val, exc_tb = sys.exc_info()
        return self._exception(exc_type, exc_val, exc_tb, message, a, kw)

    def _exception(self, exc_type, exc_val, exc_tb, message, fargs, data):
        exc_type = exc_type or DefaultException

        # have to capture the time now in case the on_exception sinks
        # take their sweet time
        etime = time.time()
        exc_info = ExceptionInfo.from_exc_info(exc_type, exc_val, exc_tb)
        if not message:
            cp = exc_info.tb_info.frames[-1]
            t = "%s raised exception: "
            exc_repr = "%s(%r)"
            errno = getattr(exc_val, 'errno', None)
            if errno and str(errno) not in exc_repr:
                t += exc_repr + ' (errno %s)' % exc_val.errno
            else:
                t += exc_repr
            t += " from %s on line %s of file '%s'"
            if self.data_map:
                t += ' - ({data_map_repr})'
            message = t % (self.name, exc_info.exc_type, exc_info.exc_msg,
                           cp.func_name, cp.lineno, cp.module_path)

        self.exc_event = ExceptionEvent(self, etime, message, fargs, exc_info)
        self.logger.on_exception(self.exc_event, exc_type, exc_val, exc_tb)

        return self._end('exception', message, fargs, data,
                         etime, exc_info)

    def _end(self, status, message, fargs, data, end_time=None, exc_info=None):
        self.data_map.update(data)

        if self._is_trans:
            end_time = end_time or time.time()
        else:
            if not self.begin_event:
                self.begin()
            end_time = self.begin_event.etime

        self.end_event = EndEvent(self, end_time, message,
                                  fargs, status, exc_info)

        if not self._defer_publish and self.logger:
            self.logger.on_end(self.end_event)

        return self

    def __enter__(self):
        self._is_trans = self._defer_publish = True
        self.logger.context.set_active_parent(self.logger, self)
        return self.begin()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._defer_publish = False
        if exc_type:
            try:
                self._exception(exc_type, exc_val, exc_tb,
                                message=None, fargs=(), data={})
            except Exception as e:
                note('record_exit',
                     'got %r while already handling exception %r', e, exc_val)
                pass  # TODO: still have to create end_event
        else:
            if self.end_event:
                self.logger.on_end(self.end_event)
            else:
                # now that _defer_publish=False, this will also publish
                self.success()

        self.logger.context.set_active_parent(self.logger, self.parent_record)

        if self._reraise is False:
            return True  # ignore exception
        return

    def __getitem__(self, key):
        return self.data_map[key]

    def __setitem__(self, key, value):
        self.data_map[key] = value

    def get_elapsed_time(self):
        """Simply get the amount of time that has passed since begin was
        called on this record, or 0.0 if it has not begun. This method
        has no side effects.
        """
        if self.begin_event:
            return time.time() - self.begin_event.etime
        return 0.0


# TODO: optimization strategy. if event creation starts to register on
# profiling, convert events to fixed-length tuples with empty
# dictionaries for caching lazy values. e.g.,
# ('begin', record, etime, event_id, to_unicode(raw_message), fargs, {})
#
# could also shove those as internal attrs on the record and put
# caching properties in place for the actual event objects. simplifies
# adding new fields.

class Event(object):
    _message = None

    def __getitem__(self, key):
        return self.record[key]

    def __getattr__(self, name):
        return getattr(self.record, name)

    @property
    def message(self):
        if self._message:
            return self._message

        raw_message = self.raw_message
        if raw_message is None:
            self._message = ''
        elif '{' not in raw_message:  # yay premature optimization
            self._message = raw_message
        else:
            # TODO: Formatter cache
            fmtr = SensibleMessageFormatter(raw_message, quoter=False)
            self._message = fmtr.format(self, *self.fargs)
        return self._message

    def __repr__(self):
        cn = self.__class__.__name__
        return '<%s %s %r>' % (cn, self.record_id, self.raw_message)


class BeginEvent(Event):
    status = 'begin'
    status_char = 'b'

    def __init__(self, record, etime, raw_message, fargs):
        self.record = record
        self.etime = etime
        self.event_id = next(_REC_ID_ITER)
        self.raw_message = to_unicode(raw_message)
        self.fargs = fargs


class ExceptionEvent(Event):
    status = 'exception'
    status_char = '!'

    def __init__(self, record, etime, raw_message, fargs, exc_info):
        self.record = record
        self.etime = etime
        self.event_id = next(_REC_ID_ITER)
        self.raw_message = to_unicode(raw_message)
        self.fargs = fargs
        self.exc_info = exc_info


class EndEvent(Event):
    def __init__(self, record, etime, raw_message, fargs, status,
                 exc_info=None):
        self.record = record
        self.etime = etime
        self.event_id = next(_REC_ID_ITER)
        self.raw_message = to_unicode(raw_message)
        self.fargs = fargs
        self.status = status
        self.exc_info = exc_info

    @property
    def status_char(self):
        if self.record._is_trans:
            ret = self.status[:1].upper()
        else:
            ret = self.status[:1].lower()
        return ret


class WarningEvent(Event):
    status = 'warning'
    status_char = 'W'

    def __init__(self, record, etime, raw_message, fargs):
        self.record = record
        self.etime = etime
        self.event_id = next(_REC_ID_ITER)
        self.raw_message = to_unicode(raw_message)
        self.fargs = fargs


class CommentEvent(Event):
    status = 'comment'
    status_char = '#'

    def __init__(self, record, etime, raw_message, fargs):
        self.record = record
        self.etime = etime
        self.event_id = next(_REC_ID_ITER)
        self.raw_message = to_unicode(raw_message)
        self.fargs = fargs


"""What to do on multiple begins and multiple ends?

If a record is atomic (i.e., never entered/begun), then should it fire
a logger on_begin? Leaning no.

Should the BeginEvent be created on record creation? currently its
presence is used to track whether on_begin has been called on the
Logger yet.

Things that normally change on a Record currently:

 - Status
 - Message
 - Data map

Things which are populated:

 - end_time
 - duration

"""

"""naming rationale:

* 'warn', an action verb, was chosen over 'warning' because it implies
  repeatability, where as success/failure/exception are nouns, to
  indicate singular conclusion.

"""



# -*- coding: utf-8 -*-

import math
import time

# 1-, 5-, and 15-minute periods, per Unix load average
DEFAULT_PERIODS = (60, 300, 900)
DEFAULT_INTERVAL = 5


class EWMAAccumulator(object):
    "An exponentially-weighted moving average (EWMA) accumulator"

    def __init__(self, periods=None, interval=None):
        "Takes an iterable of periods and an update interval, both in seconds."
        self._interval = float(interval or DEFAULT_INTERVAL)
        if not self._interval > 0:
            raise ValueError('interval must be greater than 0, not: %r'
                             % self._interval)

        periods = periods or DEFAULT_PERIODS
        self._rate_map = dict([(p, None) for p in periods])
        self._uncounted = 0.0
        self._last_update = time.time()
        self._force_next_update = True

    def add(self, value):
        "Adds a new value to the moving average."
        self._uncounted += value

    def _update(self):
        "Perform the actual decay and reset the counter."
        rate_map = self._rate_map
        cur_time = time.time()
        interval = cur_time - self._last_update

        for period in rate_map:
            new_rate = self._uncounted / interval
            if rate_map[period] is None:
                rate_map[period] = new_rate
            else:
                alpha = 1 - math.exp(-interval / period)
                rate_map[period] += alpha * (new_rate - rate_map[period])
        self._uncounted = 0
        self._last_update = cur_time

    def get_rates(self):
        "Conditionally updates and returns a copy of the rate map"
        if (time.time() - self._last_update) >= self._interval:
            self._update()
        elif self._force_next_update:
            self._force_next_update = False
            self._update()
        return dict(self._rate_map)

    def __repr__(self):
        cn = self.__class__.__name__
        return '<%s rates=%r>' % (cn, self.get_rates())


def _main(incr=1):
    speed_counts = [(0.5, 1), (0.2, 5), (0.5, 10)]
    print_interval = 0.2

    _ewma = EWMAAccumulator(interval=0.1)
    prev_time = prev_print_time = time.time()
    for speed, count in speed_counts:
        i = 0
        while i < count:
            time.sleep(0.05)

            cur_time = time.time()
            dur = cur_time - prev_time
            if dur > speed:
                _ewma.add(incr)
                i += 1
                prev_time = cur_time
            print_dur = cur_time - prev_print_time
            if print_dur > print_interval:
                prev_print_time = cur_time
                print(_ewma.get_rates())


if __name__ == '__main__':
    _main()



# -*- coding: utf-8 -*-
"""Implements types and functions for rendering
:class:`~lithoxyl.record.Record` instances into strings.
"""

import os
import time
import json
import socket
import datetime

from boltons.timeutils import UTC, LocalTZ
from boltons.formatutils import BaseFormatField
from boltons.formatutils import tokenize_format_str

from lithoxyl.common import EVENTS, IMPORT_TIME, MAX_LEVEL, get_level


__all__ = ['SensibleFormatter', 'SensibleSink']


DEFAULT_QUOTER = json.dumps

FIELD_MAP = {}
BUILTIN_FIELD_MAP = {}  # populated below


def register_builtin_field(field):
    register_field(field)
    BUILTIN_FIELD_MAP[field.fname] = field


def register_field(field):
    FIELD_MAP[field.fname] = field


class SensibleSink(object):
    def __init__(self, formatter=None, emitter=None, filters=None, on=EVENTS):
        # TODO: get_level for this
        events = on
        if isinstance(events, basestring):
            events = [events]
        unknown_events = [e for e in events if e not in EVENTS]
        if unknown_events:
            raise ValueError('unrecognized events: %r (must be one of %r)'
                             % (unknown_events, EVENTS))

        self._events = [e.lower() for e in events]
        self.filters = list(filters or [])
        self.formatter = formatter
        self.emitter = emitter

        if 'begin' in self._events:
            self.on_begin = self._on_begin
        if 'warn' in self._events:
            self.on_warn = self._on_warn
        if 'end' in self._events:
            self.on_end = self._on_end
        if 'comment' in self._events:
            self.on_comment = self._on_comment

    def _on_begin(self, event):
        if self.filters and not all([f.on_begin(event) for f in self.filters]):
            return
        entry = self.formatter.on_begin(event)
        return self.emitter.on_begin(event, entry)

    def _on_warn(self, event):
        if self.filters and not all([f.on_warn(event) for f in self.filters]):
            return
        entry = self.formatter.on_warn(event)
        return self.emitter.on_warn(event, entry)

    def _on_end(self, event):
        if self.filters and not all([f.on_end(event) for f in self.filters]):
            return
        entry = self.formatter.on_end(event)
        return self.emitter.on_end(event, entry)

    def _on_comment(self, event):
        if self.filters and not all([f.on_comment(event)
                                     for f in self.filters]):
            return
        entry = self.formatter.on_comment(event)
        return self.emitter.on_comment(event, entry)

    def __repr__(self):
        cn = self.__class__.__name__
        return ('<%s filters=%r formatter=%r emitter=%r>'
                % (cn, self.filters, self.formatter, self.emitter))


class SensibleFilter(object):
    def __init__(self, base=None, **kw):
        # TODO: on-bind lookup behaviors?

        base = get_level(base or MAX_LEVEL)

        self.begin_level = get_level(kw.pop('begin', base) or MAX_LEVEL)
        self.success_level = get_level(kw.pop('success', base) or MAX_LEVEL)
        self.failure_level = get_level(kw.pop('failure', base) or MAX_LEVEL)
        self.exception_level = get_level(kw.pop('exception', base)
                                         or MAX_LEVEL)
        self.warn_level = get_level(kw.pop('warn', base) or MAX_LEVEL)
        self.block_comments = kw.pop('block_comments', False)
        self.verbose_check = kw.pop('verbose_check', None)
        if not self.verbose_check:
            verbose_flag = kw.pop('verbose_flag', 'verbose')
            self.verbose_check = lambda e: e.record.data_map.get(verbose_flag)

        if kw:
            raise TypeError('got unexpected keyword arguments: %r' % kw)

    def on_begin(self, ev):
        if ev.record.level >= self.begin_level:
            return True
        elif self.verbose_check and self.verbose_check(ev):
            return True
        return False

    def on_end(self, ev):
        ret, rec, status = False, ev.record, ev.status
        if status == 'success':
            ret = rec.level >= self.success_level
        elif status == 'failure':
            ret = rec.level >= self.failure_level
        elif status == 'exception':
            ret = rec.level >= self.exception_level
        if not ret:
            if self.verbose_check and self.verbose_check(ev):
                ret = True
        return ret

    def on_warn(self, ev):
        if ev.record.level >= self.warn_level:
            return True
        elif self.verbose_check and self.verbose_check(ev):
            return True
        return False

    def on_comment(self, ev):
        return not self.block_comments


class GetterDict(dict):
    """An internal-use-only dict to enable the fetching of values from a
    :class:`~lithoxyl.record.Record`. Tries to fetch a key on a
    record, failing that, tries built-in getters, if that fails,
    returns ``None``. Exceptions raised from getters are not caught
    here.
    """
    def __init__(self, wrapped, getters):
        self.wrapped = wrapped
        self.getters = getters

    def __missing__(self, key):
        try:
            return self.wrapped[key]
        except KeyError:
            try:
                return self.getters[key](self.wrapped)
            except KeyError:
                pass
        return None


class SensibleFormatter(object):
    def __init__(self, base=None, **kwargs):
        defaulter = kwargs.pop('defaulter', None)
        quoter = kwargs.pop('quoter', None)
        extra_fields = kwargs.pop('extra_fields', None)

        for event in EVENTS:
            cur_fmt = kwargs.pop(event, base)
            if not cur_fmt:
                cur_fmt = ''
            rf = SensibleMessageFormatter(cur_fmt, extra_fields=extra_fields,
                                          quoter=quoter, defaulter=defaulter)
            setattr(self, '_' + event + '_formatter', rf)
        return

    def on_begin(self, begin_event):
        return self._begin_formatter(begin_event)

    def on_warn(self, warn_event):
        return self._warn_formatter(warn_event)

    def on_end(self, end_event):
        return self._end_formatter(end_event)

    def on_comment(self, comment_event):
        return self._comment_formatter(comment_event)


class SensibleMessageFormatter(object):
    """The basic ``Formatter`` type implements a constrained, but robust,
    microtemplating system rendering Records to strings that are both
    human-readable *and* machine-readable. This system is based on
    :class:`FormatFields <lithoxyl.fields.FormatField>`, many of which
    are built-in.

    Args:
        format_str (str): The template for the string to be rendered,
            e.g., ``"Request outcome: {status_str}"``.
        extra_fields (dict): Optionally specify a map of fields this
            Formatter should recognize in addition to the builtins.
        quoter (callable): Optionally override the default quoting
            function.
        defaulter (callable): Optionally override how the Formatter
            determines the default value for fields.

    Generally a Formatter is called only with the template string.

    >>> from lithoxyl.record import Record
    >>> fmtr = SensibleFormatter('Record status: {status_str}.')
    >>> fmtr.format_record(Record('test').success())
    'Record status: success.'

    Other types of Formatters do not need to inherit from any special
    class. A Formatter is any callable which accepts a Record/mapping,
    ``*args``, and ``**kwargs``, and returns text.
    """
    def __init__(self, format_str, **kwargs):
        extra_fields = kwargs.pop('extra_fields', None)
        quoter = kwargs.pop('quoter', None)
        defaulter = kwargs.pop('defaulter', None)
        if kwargs:
            raise TypeError('unexpected keyword arguments: %r' % kwargs.keys())
        if quoter is False:
            # disable quoting
            self.quoter = lambda field: None
        else:
            self.quoter = quoter or self._default_quoter
        if not callable(self.quoter):
            raise TypeError('expected callable or False for quoter,'
                            ' not %r' % self.quoter)
        self.defaulter = defaulter or self._default_defaulter
        if not callable(self.defaulter):
            raise TypeError('expected callable for defaulter, not %r'
                            % self.defaulter)

        self._field_map = dict(BUILTIN_FIELD_MAP)
        if extra_fields:
            extra_field_map = dict([(f.fname, f) for f in extra_fields or []])
            self._field_map.update(extra_field_map)
        self._getter_map = dict([(f.fname, f.getter)
                                 for f in self._field_map.values()])

        self.raw_format_str = format_str
        self.tokens = tokenize_format_str(format_str)
        self.default_map = {}
        self.quoter_map = {}
        for token in self.tokens:
            try:
                fspec = token.fspec
            except AttributeError:
                # not a field, just a string constant
                continue
            try:
                self.default_map[token] = self.defaulter(token)
                self.quoter_map[token] = self.quoter(token)
                if not fspec:
                    token.set_fspec(self._field_map[token.fname].fspec)
            except KeyError:
                # not a builtin field
                pass
        return

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.raw_format_str)

    def format(self, event, *args, **kwargs):
        """Render text, based on the format-string template used to construct
        the object, plus values from the following sources:

          * Positional arguments to this method (``*args``)
          * Keyword arguments to this method (``**kwargs``)
          * Default FormatFields built-in to Lithoxyl
          * Structured data stored in the Record object's ``data_map``

        """
        ret = ''
        kw_vals = GetterDict(event, self._getter_map)
        kw_vals.update(kwargs)
        q_map, d_map = self.quoter_map, self.default_map
        for t in self.tokens:
            try:
                name = t.base_name
            except AttributeError:
                ret += t  # just a string segment, moving on
                continue
            try:
                if t.is_positional:
                    seg = t.fstr.format(*args)
                else:
                    seg = t.fstr.format(**{name: kw_vals[name]})
                if q_map[t]:
                    seg = q_map[t](seg)
                ret += seg
            except Exception:
                ret += d_map[t]
        return ret

    def _default_defaulter(self, token):
        return str(token)

    def _default_quoter(self, token):
        field = self._field_map.get(token.fname)
        # if it's not a builtin field, we quote it by default
        if not field or field.quote:
            return DEFAULT_QUOTER
        else:
            return None

    __call__ = format


"""
reversability requirements:

all fields must be one or more of:

* known, fixed length
  (standard for single character fields, certain numeric fields maybe)
* unquoted but whitespace-free and not adjacent to any other unquoted field
  (standard for numbers and certain fixed-set labels)
* quoted, escaped
  (standard for longer strings that might contain whitespace)
"""


"""

SensibleLogger has two sinks:

 * (1) statistical (counter or stats aggregator)
 * (1) persistent (stream/file)

For each combination of level and status, choose whether to count or
count+log. The following matrix shows the default log level:

+------------+-------+-------+---------+
|level/status|success|failure|exception|
+------------+-------+-------+---------+
|debug       | count | count |   log   |
+------------+-------+-------+---------+
|info        | count |  log  |   log   |
+------------+-------+-------+---------+
|critical    |  log  |  log  |   log   |
+------------+-------+-------+---------+

Higher verbosity moves the spread of "log" actions diagonally up and
to the left, and lower verbosity, down and to the right.

"""

# Fields follow

"""Lithoxyl comes with many built-in format *fields* meant to be used
with the standard :class:`~lithoxyl.logger.Logger` and
:class:`~lithoxyl.record.Record`. Sinks can take advantage of these
with the :class:`~lithoxyl.sensible.SensibleFormatter` type or subtypes.
"""
# NOTE: docstring table needs slashes double escaped. Also, newline
# literals "\n" removed.

# TODO: exc_repr field


def timestamp2iso8601_noms(timestamp, local=False, with_tz=True):
    """
    with time.strftime(), one would have to do fractional
    seconds/milliseconds manually, because the timetuple used doesn't
    include data necessary to support the %f flag.

    This function is about twice as fast as datetime.strftime(),
    however. That's nothing compared to time.time()
    vs. datetime.now(), which is two orders of magnitude faster.
    """
    if with_tz:
        tformat = '%Y-%m-%dT%H:%M:%S %Z'
    else:
        tformat = '%Y-%m-%dT%H:%M:%S'
    if local:
        tstruct = time.localtime(timestamp)
    else:
        tstruct = time.gmtime(timestamp)
    return time.strftime(tformat, tstruct)


def timestamp2iso8601(timestamp, local=False, with_tz=True, tformat=None):
    if with_tz:
        tformat = tformat or '%Y-%m-%dT%H:%M:%S.%f%z'
    else:
        tformat = tformat or '%Y-%m-%dT%H:%M:%S.%f'
    if local:
        dt = datetime.datetime.fromtimestamp(timestamp, tz=LocalTZ)
    else:
        dt = datetime.datetime.fromtimestamp(timestamp, tz=UTC)
    return dt.strftime(tformat)


class SensibleField(BaseFormatField):
    """Fields specify whether or not they should be *quoted* (i.e.,
    whether or not values will contain whitespace or other
    delimiters), but not the exact method for their quoting. That
    aspect is reserved for the Formatter.
    """
    def __init__(self, fname, fspec='s', getter=None, **kwargs):
        quote = kwargs.pop('quote', None)
        if kwargs:
            raise TypeError('unexpected keyword arguments: %r' % kwargs.keys())
        super(SensibleField, self).__init__(fname, fspec)
        self.getter = getter
        if quote is None:
            # numeric fields should appear without quotes
            numeric = issubclass(self.type_func, (int, float))
            quote = not numeric
        self.quote = quote


def duration_auto(event):
    duration = event.record.duration
    if duration < 0.001:
        return '%.3fus' % (duration * 1e6)
    if duration < 1.0:
        return '%.3fms' % (duration * 1e3)
    return '%.4fs' % duration


# default, fmt_specs
_SF = SensibleField
BASIC_FIELDS = [_SF('logger_name', 's', lambda e: e.record.logger.name),
                _SF('logger_id', 'd', lambda e: e.record.logger.logger_id),
                _SF('record_name', 's', lambda e: e.record.name),
                _SF('record_id', 'd', lambda e: e.record_id),
                _SF('status_str', 's', lambda e: e.status, quote=False),
                _SF('status_char', 's', lambda e: e.status_char, quote=False),
                _SF('warn_char', 's', lambda e: e.warn_char, quote=False),  # TODO
                _SF('level_name', 's', lambda e: e.level_name, quote=False),
                _SF('data_map', 's', lambda e: json.dumps(e.record.data_map, sort_keys=True), quote=False),
                _SF('data_map_repr', 's', lambda e: repr(e.record.data_map), quote=False),
                _SF('level_name_upper', 's', lambda e: e.level_name.upper(), quote=False),
                _SF('level_char', 's', lambda e: e.level_name.upper()[0], quote=False),
                _SF('level_number', 'd', lambda e: e.level._value),
                _SF('begin_message', 's', lambda e: e.begin_event.message),
                _SF('begin_message_raw', 's', lambda e: e.begin_event.raw_message),
                _SF('end_message', 's', lambda e: e.record.end_event.message),
                _SF('end_message_raw', 's', lambda e: e.record.end_event.raw_message),
                _SF('begin_timestamp', '.14g', lambda e: e.record.begin_time),
                _SF('end_timestamp', '.14g', lambda e: e.record.end_time),
                _SF('duration_s', '.3f', lambda e: e.record.duration),
                _SF('duration_ms', '.3f', lambda e: e.record.duration * 1e3),
                _SF('duration_us', '.3f', lambda e: e.record.duration * 1e6),
                _SF('duration_auto', '>9s', duration_auto, quote=False),
                _SF('module_name', 's', lambda e: e.callpoint.module_name),
                _SF('module_path', 's', lambda e: e.callpoint.module_path),
                _SF('func_name', 's', lambda e: e.callpoint.func_name, quote=False),
                _SF('line_number', 'd', lambda e: e.callpoint.lineno),
                _SF('exc_type', 's', lambda e: e.exc_info.exc_type, quote=False),
                _SF('exc_message', 's', lambda e: e.exc_info.exc_msg),
                _SF('exc_tb_str', 's', lambda e: str(e.exc_info.tb_info)),
                _SF('exc_tb_list', 's', lambda e: e.exc_info.tb_info.frames),
                _SF('process_id', 'd', lambda e: os.getpid())]

# ISO8601 and variants. combinations of:
#   * begin/end
#   * UTC/Local
#   * with/without milliseconds
#   * with/without timezone (_noms variants have textual timezone)
ISO8601_FIELDS = [
    _SF('iso_begin', 's',
        lambda e: timestamp2iso8601(e.record.begin_event.etime)),
    _SF('iso_end', 's',
        lambda e: timestamp2iso8601(e.record.end_event.etime)),
    _SF('iso_begin_notz', 's',
        lambda e: timestamp2iso8601(e.record.begin_event.etime,
                                    with_tz=False)),
    _SF('iso_end_notz', 's',
        lambda e: timestamp2iso8601(e.record.end_event.etime,
                                    with_tz=False)),
    _SF('iso_begin_local', 's',
        lambda e: timestamp2iso8601(e.record.begin_event.etime,
                                    local=True)),
    _SF('iso_end_local', 's',
        lambda e: timestamp2iso8601(e.record.end_event.etime,
                                    local=True)),
    _SF('iso_begin_local_notz', 's',
        lambda e: timestamp2iso8601(e.record.begin_event.etime,
                                    local=True, with_tz=False)),
    _SF('iso_end_local_notz', 's',
        lambda e: timestamp2iso8601(e.record.end_event.etime,
                                    local=True, with_tz=False)),
    _SF('iso_begin_local_noms', 's',
        lambda e: timestamp2iso8601_noms(e.record.begin_event.etime,
                                         local=True)),
    _SF('iso_end_local_noms', 's',
        lambda e: timestamp2iso8601_noms(e.record.end_event.etime,
                                         local=True)),
    _SF('iso_begin_local_noms_notz', 's',
        lambda e: timestamp2iso8601_noms(e.record.begin_event.etime,
                                         local=True, with_tz=False)),
    _SF('iso_end_local_noms_notz', 's',
        lambda e: timestamp2iso8601_noms(e.record.end_event.etime,
                                         local=True, with_tz=False))]

# using the T separator means no whitespace and thus no quoting
for f in ISO8601_FIELDS:
    f.quote = False


DELTA_FIELDS = [
    _SF('import_delta_s', '0.6f', lambda e: e.etime - IMPORT_TIME),
    _SF('import_delta_ms', '0.4f', lambda e: (e.etime - IMPORT_TIME) * 1000)]


PARENT_DEPTH_INDENT = '   '


PARENT_FIELDS = [
    _SF('parent_depth', 'd', lambda e: e.record.parent_depth),
    _SF('parent_depth_indent', 's',
        lambda e: e.record.parent_depth * PARENT_DEPTH_INDENT,
        quote=False)]

import binascii

_GUID_SALT = '-'.join([str(os.getpid()),
                       socket.gethostname() or '<nohostname>',
                       str(time.time()),
                       binascii.hexlify(os.urandom(4))])

# I'd love to use UUID.uuid4, but this is 20x faster
import hashlib

# sha1 is 20 bytes. 12 bytes (96 bits) means that there's 1 in 2^32
# chance of a collision after 2^64 messages.


def _get_id_guid(id_int):
    return hashlib.sha1(_GUID_SALT + str(id_int)).hexdigest()[:12]


# TODO: fallback to UUID if hashlib isn't available

for f in BASIC_FIELDS:
    register_builtin_field(f)
for f in PARENT_FIELDS:
    register_builtin_field(f)
for f in ISO8601_FIELDS:
    register_builtin_field(f)
for f in DELTA_FIELDS:
    register_builtin_field(f)

del f



"""\
statsutils
~~~~~~~~~~

Provide some really basic stats functions, such as mean/average,
median, variance, the basics. A couple more advanced measures, such as
skewness, kurtosis, and median absolute deviation are also provided,
but if a project needs any more than the functions below, I'd strongly
recommend investigating ``scipy.stats`` or some other NumPy-based
statistics package.

(Note that all functions default to 0.0 when passed no values, which,
while not totally accurate, has usually worked out fine for basic
usage.)

A note on statistical moments
-----------------------------

Most people are probably familiar with the concept of the "mean" or
"average", which gives a sort of quantitiative "middle-ish" value by
which a sample can be can be generalized. But the mean is just the
first of four ways a sample or distribution can be described:

Statistical moments and what they generally describe:

1. Average - theoretical middle value
2. Variance - width of value dispersion
3. Skewness - symmetry of distribution
4. Kurtosis - "peakiness" or long-tailed-ness

For more information it's probably best to read a bit of Wikipedia:

http://en.wikipedia.org/wiki/Moment_(mathematics)

Keep in mind that while these moments can give a bit more insight into
the shape and distribution of data, they do not guarantee a complete
picture. Wildly different datasets can have the same values for all
four moments, so generalize wisely. (And read up on robust statistics
and scipy.stats)
"""

__all__ = ['mean', 'median', 'variance', 'std_dev', 'rel_std_dev',
           'median_abs_dev', 'skewness', 'kurtosis', 'trim_sample']


def mean(vals):
    """
    The arithmetic mean, or "average". Sum of the values divided by
    the number of values.

    >>> mean([1, 2, 3])
    2.0
    >>> mean(range(97))
    48.0
    >>> mean(range(96) + [1066])  # 1066 is an arbitrary outlier
    58.0
    """
    if vals:
        return sum(vals, 0.0) / len(vals)
    else:
        return 0.0


def median(vals):
    """
    The median is either the middle value or the average of the two
    middle values of a sample. Compared to the mean, it's generally
    more resilient to the presence of outliers in the sample.

    >>> median([1, 2, 3])
    2
    >>> median(range(97))
    48
    >>> median(range(96) + [1066])  # 1066 is an arbitrary outlier
    48
    """
    if not vals:
        return 0.0
    sorted_vals, size = sorted(vals), len(vals)
    mid = size // 2  # aka floor division
    if size % 2 == 1:
        return sorted_vals[mid]
    else:
        return (sorted_vals[mid - 1] + sorted_vals[mid]) / 2.0


def variance(vals):
    """\
    Variance is the average of the squares of the difference between
    each value and the mean.

    >>> variance(range(97))
    784.0
    """
    return mean(_pow_diff(vals, 2))


def std_dev(vals):
    """\
    Standard deviation. Square root of the variance.

    >>> std_dev(range(97))
    28.0
    """
    return variance(vals) ** 0.5


def median_abs_dev(vals):
    """\
    Median Absolute Deviation is a robust measure of statistical
    dispersion: http://en.wikipedia.org/wiki/Median_absolute_deviation

    >>> median_abs_dev(range(97))
    24.0
    """
    sorted_vals = sorted(vals)
    x = float(median(sorted_vals))
    return median([abs(x - v) for v in sorted_vals])


def rel_std_dev(vals):
    """\
    Standard deviation divided by the absolute value of the average.

    http://en.wikipedia.org/wiki/Relative_standard_deviation

    >>> round(rel_std_dev(range(97)), 3)
    0.583
    """
    abs_mean = abs(mean(vals))
    if abs_mean:
        return std_dev(vals) / abs_mean
    else:
        return 0.0


def skewness(vals):
    """\
    Indicates the asymmetry of a curve. Positive values mean the bulk
    of the values are on the left side of the average and vice versa.

    http://en.wikipedia.org/wiki/Skewness

    See the module docstring for more about statistical moments.

    >>> skewness(range(97))  # symmetrical around 48.0
    0.0
    >>> left_skewed = skewness(range(97) + range(10))
    >>> right_skewed = skewness(range(97) + range(87, 97))
    >>> round(left_skewed, 3), round(right_skewed, 3)
    (0.114, -0.114)
    """
    s_dev = std_dev(vals)
    if len(vals) > 1 and s_dev > 0:
        return (sum(_pow_diff(vals, 3)) /
                float((len(vals) - 1) * (s_dev ** 3)))
    else:
        return 0.0


def kurtosis(vals):
    """\
    Indicates how much data is in the tails of the distribution. The
    result is always positive, with the normal "bell-curve"
    distribution having a kurtosis of 3.

    http://en.wikipedia.org/wiki/Kurtosis

    See the module docstring for more about statistical moments.

    >>> kurtosis(range(9))
    1.99125

    With a kurtosis of 1.99125, [0, 1, 2, 3, 4, 5, 6, 7, 8] is more
    centrally distributed than the normal curve.
    """
    s_dev = std_dev(vals)
    if len(vals) > 1 and s_dev > 0:
        return (sum(_pow_diff(vals, 4)) /
                float((len(vals) - 1) * (s_dev ** 4)))
    else:
        return 0.0


def trim_sample(vals, trim=0.25):
    """
    A utility function used to cut a proportion of values off each end
    of a list of values. When sorted, this has the effect of limiting
    the effect of outliers.

    >>> trim_sample(range(12), 0.25)
    [3, 4, 5, 6, 7, 8]
    """
    if trim > 0.0:
        trim = float(trim)
        size = len(vals)
        size_diff = int(size * trim)
        vals = vals[size_diff:-size_diff]
    return vals


def _pow_diff(vals, power):
    """
    A utility function used for calculating statistical moments.
    """
    m = mean(vals)
    return [(v - m) ** power for v in vals]



# -*- coding: utf-8 -*-
"""The :class:`Logger` is the application developer's primary
interface to using Lithoxyl. It is used to conveniently create
:class:`Records <Record>` and publish them to :class:`sinks <Sink>`.

"""

import sys
import time
import itertools
from collections import deque
from threading import RLock

from boltons.funcutils import wraps

from lithoxyl.context import get_context
from lithoxyl.common import DEBUG, INFO, CRITICAL
from lithoxyl.record import Record, BeginEvent, EndEvent, CommentEvent


QUEUE_LIMIT = 10000
_LOG_ID_ITER = itertools.count()


def _get_previous_frame(frame):
    try:
        return frame.f_back
    except AttributeError:
        raise ValueError('reached topmost frame in stack')


# TODO: should all sys._getframes be converted to use this?
# TODO: could precalculate offsets based on which methods are overridden
# TODO: also precalculation could happen in a metaclass
def get_frame_excluding_subtypes(target_type, offset=0):
    """
    `offset` is the number of additional frames to look up after
    reaching the outside of a class (in the event of a factory
    function or some such.
    """
    frame = sys._getframe(1)
    args = frame.f_code.co_varnames[:frame.f_code.co_argcount]
    first_arg_name = args[0] if args else ''
    i = 0
    while 1:
        i += 1
        first_arg = frame.f_locals.get(first_arg_name)
        if i > 10000:
            raise ValueError('could not get frame')
        if isinstance(first_arg, target_type):
            frame = _get_previous_frame(frame)
        elif isinstance(first_arg, type) and issubclass(first_arg, target_type):
            frame = _get_previous_frame(frame)
        else:
            break
    for i in xrange(offset):
        frame = _get_previous_frame(frame)
    return frame


class Logger(object):
    """The ``Logger`` is one of three core Lithoxyl types, and the main
    entrypoint to creating :class:`~lithoxyl.record.Record` instances,
    and publishing those :term:`records <record>` to :term:`sinks
    <sink>`.

    Args:
        name (str): Name of this Logger.
        sinks (list): A list of :term:`sink` objects to be attached to
            the Logger. Defaults to ``[]``. Sinks can be added later
            with :meth:`Logger.add_sink`.
        module (str): Name of the module where the new Logger instance
            will be stored.  Defaults to the module of the caller.

    Most Logger methods and attributes fal into three categories:
    :class:`~lithoxyl.record.Record` creation, Sink registration, and
    Event handling.
    """

    record_type = Record
    "Override *record_type* in subtypes for custom Record behavior."

    def __init__(self, name, sinks=None, **kwargs):
        self.logger_id = next(_LOG_ID_ITER)

        self.context = kwargs.pop('context', None) or get_context()
        self.context.add_logger(self)
        # TODO context-configurable
        self.event_queue = deque(maxlen=QUEUE_LIMIT)
        self.async_mode = kwargs.pop('async', self.context.async_mode)
        self.async_lock = RLock()
        self.preflush_hooks = []
        self.last_flush = time.time()

        self.module = kwargs.pop('module', None)
        self._module_offset = kwargs.pop('module_offset', 0)
        if self.module is None:
            frame = get_frame_excluding_subtypes(target_type=Logger,
                                                 offset=self._module_offset)
            self.module = frame.f_globals.get('__name__', '<module>')
        if kwargs:
            raise TypeError('unexpected keyword arguments: %r' % kwargs)
        self.name = name or self.module
        self.set_sinks(sinks)

    def set_async(self, enabled):
        self.async_mode = enabled

    def flush(self):
        # only one flush allowed to run at a time
        # ensures that records are delivered to sinks in order
        with self.async_lock:
            for preflush_hook in self.preflush_hooks:
                try:
                    preflush_hook(self)
                except Exception as e:
                    self.context.note('preflush', 'hook %r got exception %r',
                                      preflush_hook, e)
            queue = self.event_queue
            while queue:
                ev_type, ev = queue.popleft()
                if ev_type == 'begin':
                    for begin_hook in self._begin_hooks:
                        begin_hook(ev)
                elif ev_type == 'end':
                    for end_hook in self._end_hooks:
                        end_hook(ev)
                elif ev_type == 'warn':
                    for warn_hook in self._warn_hooks:
                        warn_hook(ev)
                elif ev_type == 'comment':
                    for comment_hook in self._comment_hooks:
                        comment_hook(ev)
                else:
                    self.context.note('flush', 'unknown event type: %r %r',
                                      ev_type, ev)
        self.last_flush = time.time()
        return

    @property
    def sinks(self):
        """A copy of all sinks set on this Logger.
        Set sinks with :meth:`Logger.set_sinks`.
        """
        return list(self._all_sinks)

    def set_sinks(self, sinks):
        "Replace this Logger's sinks with *sinks*."
        sinks = sinks or []
        self._all_sinks = []
        self._begin_hooks = []
        self._warn_hooks = []
        self._end_hooks = []
        self._exc_hooks = []
        self._comment_hooks = []
        for s in sinks:
            self.add_sink(s)

    def clear_sinks(self):
        "Clear this Logger's sinks."
        self.set_sinks([])

    def add_sink(self, sink):
        """Add *sink* to this Logger's sinks. Does nothing if *sink* is
        already in this Logger's sinks.
        """
        # TODO: check signatures?
        if sink in self._all_sinks:
            return
        begin_hook = getattr(sink, 'on_begin', None)
        if callable(begin_hook):
            self._begin_hooks.append(begin_hook)
        warn_hook = getattr(sink, 'on_warn', None)
        if callable(warn_hook):
            self._warn_hooks.append(warn_hook)
        end_hook = getattr(sink, 'on_end', None)
        if callable(end_hook):
            self._end_hooks.append(end_hook)
        exc_hook = getattr(sink, 'on_exception', None)
        if callable(exc_hook):
            self._exc_hooks.append(exc_hook)
        comment_hook = getattr(sink, 'on_comment', None)
        if callable(comment_hook):
            self._comment_hooks.append(comment_hook)
        # TODO: also pull flush methods?
        self._all_sinks.append(sink)

    def on_end(self, end_event):
        "Publish *end_event* to all sinks with ``on_end()`` hooks."
        if self.async_mode:
            self.event_queue.append(('end', end_event))
        else:
            for end_hook in self._end_hooks:
                end_hook(end_event)
        return

    def on_begin(self, begin_event):
        "Publish *begin_event* to all sinks with ``on_begin()`` hooks."
        if self.async_mode:
            self.event_queue.append(('begin', begin_event))
        else:
            for begin_hook in self._begin_hooks:
                begin_hook(begin_event)
        return

    def on_warn(self, warn_event):
        "Publish *warn_event* to all sinks with ``on_warn()`` hooks."
        if self.async_mode:
            self.event_queue.append(('warn', warn_event))
        else:
            for warn_hook in self._warn_hooks:
                warn_hook(warn_event)
        return

    def on_exception(self, exc_event, exc_type, exc_obj, exc_tb):
        "Publish *exc_event* to all sinks with ``on_exception()`` hooks."
        # async handling doesn't make sense here
        for exc_hook in self._exc_hooks:
            exc_hook(exc_event, exc_type, exc_obj, exc_tb)
        return

    def comment(self, message, *a, **kw):
        # comments are not enterable, they're not returned
        rec_type = self.record_type
        rec = rec_type(logger=self, level=CRITICAL, name='comment', data=kw,
                       parent=kw.pop('parent_record', None))
        cur_time = time.time()
        rec.begin_event = BeginEvent(rec, cur_time, 'comment', ())
        rec.end_event = EndEvent(rec, cur_time,
                                 'comment', (), 'success')
        event = CommentEvent(rec, cur_time, message, a)
        if self.async_mode:
            self.event_queue.append(('comment', event))
        else:
            for comment_hook in self._comment_hooks:
                comment_hook(event)
        return

    def debug(self, record_name, **kw):
        "Returns a new :data:`DEBUG`-level :class:`Record` named *name*."
        return self.record_type(logger=self, level=DEBUG, name=record_name,
                                data=kw, reraise=kw.pop('reraise', None),
                                parent=kw.pop('parent_record', None),
                                frame=sys._getframe(1))

    def info(self, record_name, **kw):
        "Returns a new :data:`INFO`-level :class:`Record` named *name*."
        return self.record_type(logger=self, level=INFO, name=record_name,
                                data=kw, reraise=kw.pop('reraise', None),
                                parent=kw.pop('parent_record', None),
                                frame=sys._getframe(1))

    def critical(self, record_name, **kw):
        "Returns a new :data:`CRITICAL`-level :class:`Record` named *name*."
        return self.record_type(logger=self, level=CRITICAL, name=record_name,
                                data=kw, reraise=kw.pop('reraise', None),
                                parent=kw.pop('parent_record', None),
                                frame=sys._getframe(1))

    def record(self, level, record_name, **kw):
        "Return a new :class:`Record` named *name* classified as *level*."
        return self.record_type(logger=self, level=level, name=record_name,
                                data=kw, reraise=kw.pop('reraise', None),
                                parent=kw.pop('parent_record', None),
                                frame=sys._getframe(1))

    def wrap(self, level, record_name=None,
             inject_as=None, enable_wrap=True, **kw):

        record_kwargs = kw

        def record_wrapper(func_to_log,
                           _enable=enable_wrap,
                           _name=record_name,
                           _record_kwargs=record_kwargs):
            if not _enable:
                return func_to_log
            if _name is None:  # wooo nonlocal
                _name = func_to_log.__name__

            @wraps(func_to_log, injected=inject_as)
            def logged_func(*a, **kw):
                rec = self.record(level, _name, **record_kwargs)
                if inject_as:
                    kw[inject_as] = rec
                with rec:
                    return func_to_log(*a, **kw)

            wrapping_info = (self, level, record_name, func_to_log)
            logged_func.__lithoxyl_wrapped__ = wrapping_info

            return logged_func

        return record_wrapper

    def __repr__(self):
        cn = self.__class__.__name__
        try:
            return '<%s name=%r sinks=%r>' % (cn, self.name, self.sinks)
        except Exception:
            return object.__repr__(self)




import os
import time
import threading

# TODO: what to do in systems without threading?
# TODO: signal actor for kicks?

DEFAULT_INTERVAL = 200  # milliseconds


class IntervalThreadActor(object):
    """Manages a thread that calls a *task* function and waits
    *interval* milliseconds before calling the function again.

    Failures and exceptions do not halt this loop.

    Args:
      task (callable): Function to call periodically. Takes no arguments.
      interval (number): Milliseconds to wait before next call of `process`.

    """
    def __init__(self, task, interval=DEFAULT_INTERVAL, **kwargs):
        self.task = task
        if not callable(task):
            raise ValueError('expected callable for task, not %r' % task)

        self._thread = None
        self._stopping = threading.Event()
        self._pid = None

        if interval is None:
            interval = DEFAULT_INTERVAL
        self.interval = self._orig_interval = float(interval)
        max_interval = kwargs.pop('max_interval', None)
        self.max_interval = float(max_interval or interval * 8)
        self._daemonize_thread = kwargs.pop('daemonize_thread', True)
        self._note = kwargs.pop('note', None)
        if self._note and not callable(self._note):
            raise ValueError('expected callable for note, not %r' % self._note)
        if kwargs:
            raise TypeError('unexpected keyword arguments: %r' % kwargs.keys())

        self._run_start_time = 0
        self._task_call_time = 0
        self._task_call_count = 0

    def get_stats(self):
        ret = {'run_start_time': self._run_start_time,
               'task_call_time': self._task_call_time,
               'task_call_count': self._task_call_count}
        return ret

    def is_alive(self):
        return self._thread and self._thread.is_alive()

    def start(self):
        if self.is_alive():
            return
        # os.getpid compare allows restarting after forks
        if os.getpid() == self._pid and self._stopping.is_set():
            # alive and stopping
            raise RuntimeError('expected caller to wait on join'
                               ' before calling start again')
        self._pid = os.getpid()
        self._stopping.clear()
        self._thread = threading.Thread(target=self._run)
        self._thread.daemon = self._daemonize_thread
        self._thread.start()
        return

    def stop(self):
        "actually, 'start_stopping' is more accurate"
        # TODO: what if we fork and the _run hasn't finished stopping,
        # so stopping is still set, but the thread is not alive. need
        # to reset the stopping state, or raise an error because the
        # data might not be clean.
        if self.is_alive():
            self._stopping.set()
        else:
            self._stopping.clear()
        return

    def join(self, timeout=None):
        if not self._thread:
            raise RuntimeError('actor must be started before it can be joined')
        self._thread.join(timeout=timeout)
        ret = self.is_alive()
        if not ret:
            self._stopping.clear()
        return ret

    def note(self, name, message, *a, **kw):
        if self._note:
            name = 'actor_' + str(name)
            self._note(name, message)
        return

    def _run(self):
        self._run_start_time = time.time()
        # TODO: start delay/jitter?
        try:
            while not self._stopping.is_set():
                self._task_call_count += 1
                cur_start_time = time.time()
                try:
                    self.task()
                except (SystemExit, KeyboardInterrupt):
                    if not self._daemonize_thread:
                        raise
                except Exception as e:
                    self.note('task_exception', '%s - task() (%r) raised: %r'
                              % (time.time(), self.task, e))
                    self.interval = min(self.interval * 2, self.max_interval)
                else:
                    decrement = (self.max_interval - self._orig_interval) / 8
                    self.interval = max(self.interval - decrement,
                                        self._orig_interval)
                cur_duration = time.time() - cur_start_time
                self._task_call_time += cur_duration
                interval_seconds = self.interval / 1000.0
                self._stopping.wait(interval_seconds)
        finally:
            self._stopping.clear()
        return



"""A systematic approach to application instrumentation, including
logging, semantic profiling, and statistics collection. Very
lightweight, very Pythonic.

(c) 2016 by Mahmoud Hashemi.
BSD-licensed, see LICENSE for more details.

"""

import sys
from setuptools import setup, find_packages


__author__ = 'Mahmoud Hashemi'
__version__ = '0.3.0'
__contact__ = 'mahmoudrhashemi@gmail.com'
__url__ = 'https://github.com/mahmoud/lithoxyl'
__license__ = 'BSD'

desc = ('A systematic approach to logging, profiling, and statistics'
        'collection. Very lightweight, very Pythonic.')


if sys.version_info < (2,6):
    raise NotImplementedError("Sorry, lithoxyl only supports Python >=2.6")


if sys.version_info >= (3,):
    raise NotImplementedError("lithoxyl Python 3 support en route to your location")


setup(name='lithoxyl',
      version=__version__,
      description=desc,
      long_description=__doc__,
      author=__author__,
      author_email=__contact__,
      url=__url__,
      packages=find_packages(),
      install_requires=['boltons>=16.5.0'],
      include_package_data=True,
      zip_safe=False,
      license=__license__,
      platforms='any',
      classifiers=[
          'Intended Audience :: Developers',
          'Topic :: System :: Logging',
          'Topic :: Utilities',
          'Topic :: Software Development :: Libraries :: Application Frameworks',
          'License :: OSI Approved :: BSD License',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7', ]
      )



# -*- coding: utf-8 -*-

from lithoxyl.context import get_context, set_context

from lithoxyl.logger import Logger, DEBUG, INFO, CRITICAL
from lithoxyl.emitters import StreamEmitter

from lithoxyl.sensible import (SensibleSink,
                               SensibleFilter,
                               SensibleFormatter,
                               SensibleMessageFormatter)
from lithoxyl.sinks import AggregateSink


from boltons.formatutils import DeferredValue
from boltons.tbutils import TracebackInfo, ExceptionInfo, Callpoint



# -*- coding: utf-8 -*-

import syslog

from lithoxyl.common import DEBUG, INFO, CRITICAL, get_level


class SyslogEmitter(object):

    default_priority_map = {DEBUG: {'success': syslog.LOG_DEBUG,
                                    'failure': syslog.LOG_INFO,
                                    'warn': syslog.LOG_INFO,
                                    'exception': syslog.LOG_NOTICE},
                            INFO: {'success': syslog.LOG_INFO,
                                   'failure': syslog.LOG_NOTICE,
                                   'warn': syslog.LOG_NOTICE,
                                   'exception': syslog.LOG_WARNING},
                            CRITICAL: {'success': syslog.LOG_NOTICE,
                                       'failure': syslog.LOG_WARNING,
                                       'warn': syslog.LOG_WARNING,
                                       'exception': syslog.LOG_ERR}}

    def __init__(self, ident, priority_map=None,
                 options=syslog.LOG_PID, facility=syslog.LOG_USER):
        self.ident = ident
        self.options = options
        self.facility = facility

        # TODO merge and resolve
        self.priority_map = priority_map or self.default_priority_map

        self.syslog_conn = syslog.openlog(ident, options, facility)

    def _get_syslog_priority(self, event_name, record):
        level = get_level(record.level)

        if event_name == 'warn':
            status = 'warn'
        elif event_name == 'begin':
            status = 'begin'
        else:
            status = record.status
        return self.priority_map[level][status]

    def on_begin(self, begin_event, entry):
        priority = self._get_syslog_priority('begin', begin_event)
        syslog.syslog(priority, entry)

    def on_warn(self, warn_event, entry):
        priority = self._get_syslog_priority('warn', warn_event)
        syslog.syslog(priority, entry)

    def on_end(self, end_event, entry):
        priority = self._get_syslog_priority('end', end_event)
        syslog.syslog(priority, entry)



# -*- coding: utf-8 -*-

import sys
import time
import json

from lithoxyl.sinks import AggregateSink
from lithoxyl.logger import Logger


def _get_logger():
    acc = AggregateSink()
    return Logger('test_logger', [acc])


def do_debug_trans(logger):
    with logger.debug('hi') as t:
        time.sleep(0.01)
        t.success('yay')
    return t


def test_logger_success(trans_count=2):
    logger = _get_logger()
    for i in range(trans_count):
        do_debug_trans(logger)
    assert len(logger.sinks[0].end_events) == trans_count
    agg_sink_repr = repr(logger.sinks[0])
    assert 'begins=' in agg_sink_repr
    assert 'begins=0' not in agg_sink_repr


def test_callpoint_info():
    log = Logger('test_logger', [])
    t = do_debug_trans(log)
    assert t.callpoint.module_name == __name__
    assert t.callpoint.module_path.endswith(__file__)
    assert t.callpoint.func_name == 'do_debug_trans'
    assert t.callpoint.lineno > 0
    assert t.callpoint.lasti > 0
    assert repr(t)


def test_reraise_false():
    logger = _get_logger()
    with logger.debug('hi', reraise=False) as t:
        x
    assert logger.sinks[0].end_events[0].status == 'exception'


def test_reraise_true():
    logger = _get_logger()
    try:
        with logger.debug('hi', reraise=True) as t:
            y
    except NameError:
        assert True
    else:
        assert False, 'should have reraised NameError'


_MSG_ATTRS = ('name', 'level_name', 'status', 'message',
              'begin_time', 'end_time', 'duration')


class SimpleStructuredFileSink(object):
    def __init__(self, fileobj=None):
        self.fileobj = fileobj or sys.stdout

    def on_end(self, event):
        msg_data = dict(event.data_map)
        for attr in _MSG_ATTRS:
            msg_data[attr] = getattr(event, attr, None)
        json_str = json.dumps(msg_data, sort_keys=True)
        self.fileobj.write(json_str)
        self.fileobj.write('\n')


def test_structured(trans_count=5):
    acc = SimpleStructuredFileSink()
    log = Logger('test_logger', [acc])
    for i in range(trans_count):
        do_debug_trans(log)
